## @package freesasa
# @author Simon Mitternacht
# @copyright [MIT License](md_license.html)
#
# @brief FreeSASA Python interface.
#
# A minimal program would be something like
#
# ~~~{.py}
#     structure = freesasa.Structure("1abc.pdb")
#     result = freesasa.calc(structure)
#     print "SASA = %f" % result.totalArea()
# ~~~
#
# See documentation of the classes and functions for how to customize behavior.
#

from libc.stdio cimport FILE, fopen, fclose
from libc.stdlib cimport free, realloc, malloc
from libc.string cimport memcpy
from cfreesasa cimport *

## Used to specify the algorithm by Shrake & Rupley
ShrakeRupley = 'ShrakeRupley'

## Used to specify the algorithm by Lee & Richards
LeeRichards = 'LeeRichards'

## Used for classification
polar = 'Polar'

## Used for classification
apolar = 'Apolar'

## int: Suppress all warnings and errors (used by setVerbosity())
silent = FREESASA_V_SILENT

## int: Suppress all warnings but not errors (used by setVerbosity())
nowarnings = FREESASA_V_NOWARNINGS

## int: Normal verbosity (used by setVerbosity())
normal = FREESASA_V_NORMAL

## int: Print debug messages (used by setVerbosity())
debug = FREESASA_V_DEBUG

## The default values for calculation parameters
defaultParameters = {
     'algorithm'    : LeeRichards, 
     'probe-radius' : freesasa_default_parameters.probe_radius,
     'n-points'     : freesasa_default_parameters.shrake_rupley_n_points,
     'n-slices'     : freesasa_default_parameters.lee_richards_n_slices,
     'n-threads'    : freesasa_default_parameters.n_threads 
}

## Stores parameter values to be used by calculation.
#   
#  Wraps the C struct freesasa_parameters.
cdef class Parameters:

      cdef freesasa_parameters _c_param
      ## Initializes Parameters object.
      #      
      #  @param param (dict) optional argument to specify parameter-values, modeled after
      #            ::defaultParameters
      #  @exception AssertionError invalid parameter values supplied
      #  @see defaultParameters
      def __init__(self,param=None):
   
            self._c_param = freesasa_default_parameters
            if param != None:
                  if 'algorithm' in param:    self.setAlgorithm(param['algorithm'])
                  if 'probe-radius' in param: self.setProbeRadius(param['probe-radius'])
                  if 'n-points' in param:     self.setNPoints(param['n-points'])
                  if 'n-slices' in param:     self.setNSlices(param['n-slices'])
                  if 'n-threads' in param:    self.setNThreads(param['n-threads'])
                  unknownKeys = []
                  for key in param:
                        if not key in defaultParameters:
                              unknownKeys.append(key)
                  if len(unknownKeys) > 0:
                        raise AssertionError('Key(s): ',unknownKeys,', unknown');

      ## Set algorithm.
      #
      #  @param alg (str) algorithm name, only allowed values are ::ShrakeRupley and ::LeeRichards
      #  @exception AssertionError unknown algorithm specified
      def setAlgorithm(self,alg):
            if alg == ShrakeRupley:
                  self._c_param.alg = FREESASA_SHRAKE_RUPLEY
            elif alg == LeeRichards:
                  self._c_param.alg = FREESASA_LEE_RICHARDS
            else:
                  raise AssertionError("Algorithm '%s' is unknown" % alg)

      ## Get algorithm.
      #     
      #  @return Name of algorithm
      def algorithm(self):
            if self._c_param.alg == FREESASA_SHRAKE_RUPLEY:
                  return ShrakeRupley
            if self._c_param.alg == FREESASA_LEE_RICHARDS:
                  return LeeRichards
            raise Exception("No algorithm specified, shouldn't be possible")

      ## Set probe radius.
      # @param r probe radius in Å (>= 0)
      # @exception AssertionError r < 0
      def setProbeRadius(self,r):
            assert(r >= 0)
            self._c_param.probe_radius = r

      ## Get probe radius.
      #  @return Probe radius in Å
      def probeRadius(self):
            return self._c_param.probe_radius

      ## Set number of test points in Shrake & Rupley algorithm.
      #  @param n (int) Number of points (> 0). 
      #  @exception AssertionError n <= 0.
      def setNPoints(self,n):
            assert(n > 0)
            self._c_param.shrake_rupley_n_points = n

      ## Get number of test points in Shrake & Rupley algorithm.
      #  @return Number of points.
      def nPoints(self):
            return self._c_param.shrake_rupley_n_points

      ## Set the number of slices per atom in Lee & Richards algorithm.
      #  @param n (int) Number of slices (> 0)
      #  @exception AssertionError n <= 0
      def setNSlices(self,n):
            assert(n> 0)
            self._c_param.lee_richards_n_slices = n

      ## Get the number of slices per atom in Lee & Richards algorithm.
      #  @return Number of slices.
      def nSlices(self):
            return self._c_param.lee_richards_n_slices

      
      ## Set the number of threads to use in calculations.
      #  @param n (int) Number of points (> 0)
      #  @exception AssertionError n <= 0
      def setNThreads(self,n):
            assert(n>0)
            self._c_param.n_threads = n

            
      ## Get the number of threads to use in calculations.
      #  @return Number of threads.
      def nThreads(self):
            return self._c_param.n_threads

      # not pretty, but only way I've found to pass pointers around
      def _get_address(self, size_t ptr2ptr):
            cdef freesasa_parameters **p = <freesasa_parameters**> ptr2ptr
            p[0] = &self._c_param

## Stores results from SASA calculation.
#  The type of object returned by calc(), not intended to be used
#  outside of that context. 
#
#  Wraps the C struct freesasa_result.
cdef class Result:
      cdef freesasa_result* _c_result

      ## The constructor
      def __cinit__ (self):
            self._c_result = NULL

      ## The destructor
      def __dealloc__(self):
            if self._c_result is not NULL:
                  freesasa_result_free(self._c_result)
            
      ## Number of atoms in the results
      #  @return Number of atoms
      def nAtoms(self):
            if self._c_result is not NULL:
                  return self._c_result.n_atoms
            return 0

      
      ## Total SASA.
      # @return The total area in Å^2.
      # @exception AssertionError If no results have been associated
      #              with the object
      def totalArea(self):
            assert(self._c_result is not NULL)
            return self._c_result.total

      ## SASA for a given atom.
      #  @param i (int) index of atom.
      #  @return SASA of atom i in Å^2.
      #  @exception AssertionError If no results have been associated
      #              with the object or if index is out of bounds
      def atomArea(self,i):
            assert(self._c_result is not NULL)
            assert(i < self._c_result.n_atoms)
            return self._c_result.sasa[i]

      def _get_address(self, size_t ptr2ptr):
            cdef freesasa_result **p = <freesasa_result**> ptr2ptr
            p[0] = self._c_result

      
## Assigns class and radius to atom by residue and atom name.
#
#  Subclasses derived from Classifier can be used to define custom
#  atomic radii and/or classes. Can also be initialized from a @ref
#  Config-file "configuration file" with a custom classifier.
#
#  Wraps a C freesasa_classifier. If initialized without arguments the
#  default classifier is used.
#
#  Derived classifiers should set the member purePython to True
#
#  Residue names should be of the format `"ALA"`,`"ARG"`, etc.
#      
#  Atom names should be of the format `"CA"`, `"N"`, etc. 
#
cdef class Classifier:
      cdef freesasa_classifier* _c_classifier
      purePython = False

      ## Constructor.
      #
      #  If no file is provided the default classifier is used. 
      #
      #  @see @ref Config-file.
      #      
      #  @param fileName Name of file with classifier configuration.
      #  @exception IOError   Problem opening/reading file
      #  @exception Exception Problem parsing provided configuration or 
      #                       initializing defaults
      def __cinit__ (self,fileName=None):
            cdef FILE *config
            self._c_classifier = NULL
            if fileName is not None:
                  config = fopen(fileName,'r')
                  if config is NULL:
                        raise IOError("File '%s' could not be opened." % fileName)
                  self._c_classifier = freesasa_classifier_from_file(config)
                  fclose(config)
                  if self._c_classifier is NULL:
                        raise Exception("Error parsing configuration in '%s'." % fileName)
            else:
                  self._c_classifier = &freesasa_default_classifier
      ## The destructor
      def __dealloc__(self):
            if self._c_classifier is not &freesasa_default_classifier:
                  freesasa_classifier_free(self._c_classifier)


      # This is used internally to determine if a Classifier wraps a C
      # classifier or not (necessary when generating structures)
      # @return Boolean
      def _isCClassifier(self):
            return not self.purePython

      ## Class of atom.
      #
      #  Depending on the configuration these classes can be
      #  anything, but typically they will be 'Polar' and 'Apolar'. 
      #  Unrecognized atoms will get the class 'Unknown'. 
      #
      #  @param residueName (str) Residue name (`"ALA"`,`"ARG"`,...).
      #  @param atomName (str) Atom name (`"CA"`,`"C"`,...).
      #  @return A string describing the class
      def classify(self,residueName,atomName):
            classIndex = freesasa_classifier_class(self._c_classifier, residueName, atomName)
            return freesasa_classifier_class2str(classIndex)

      ## Radius of atom.
      #
      #  This allows the classifier to be used to calculate the atomic
      #  radii used in calculations. Unknown atoms will get a negative
      #  radius.
      #
      #  @param residueName (str) Residue name (`"ALA"`, `"ARG"`, ...).
      #  @param atomName (str) Atom name (`"CA"`, `"C"`, ...).
      #  @return The radius in Å.
      def radius(self,residueName,atomName):
            return freesasa_classifier_radius(self._c_classifier, residueName, atomName)

      def _get_address(self, size_t ptr2ptr):
            cdef freesasa_classifier **p = <freesasa_classifier**> ptr2ptr
            p[0] = self._c_classifier

## Represents a protein structure, including its atomic radii.
#      
#  Initialized from PDB-file. Calculates atomic radii using default
#  classifier, or custom one provided as argument to initalizer
#
#  Wraps the C struct freesasa_structure.
#
#  Since it is intended to be a static structure the word 'get' is
#  omitted in the getter-functions.
cdef class Structure:
      cdef freesasa_structure* _c_structure
      ## By default ignore HETATM, Hydrogens, only use first model. For unknown atoms
      # try to guess the radius, if this fails, assign radius 0 (to
      # allow changing the radius later).
      defaultOptions = {         'hetatm' : False,
            'hydrogen' : False,
            'join-models' : False,
            'skip-unknown' : False,
            'halt-at-unknown' : False
            }

      ## Constructor
      #
      #  If PDB file is provided, the structure will be constructed
      #  based on the file. If not, this simply initializes an empty
      #  structure and the other arguments are ignored. In this case 
      #  atoms will have to be added manually using addAtom().
      #
      #  @param fileName (str) PDB file (if None empty structure generated).
      #  @param classifier An optional Classifier to calculate atomic
      #           radii, uses default if none provided
      #  @param options specify which atoms and models to include
      #  @exception IOError Problem opening/reading file.
      #  @exception Exception Problem parsing PDB file or calculating 
      #      atomic radii.
      #  @exception Exception If option 'halt-at-unknown' selected and
      #      unknown atom encountered.
      def __init__(self,fileName=None,classifier=None,
                   options = defaultOptions):

            self._c_structure = NULL
            cdef freesasa_classifier *c = NULL
            if classifier is None:
                  classifier = Classifier()
            if classifier._isCClassifier():
                  classifier._get_address(<size_t>&c)

            if fileName is None:
                  self._c_structure = freesasa_structure_new()
                  return
            cdef FILE *input
            input = fopen(fileName,'r')
            if input is NULL:
                  raise IOError("File '%s' could not be opened." % fileName)
            structure_options = Structure._get_structure_options(options)

            if not classifier._isCClassifier(): # supress warnings
                  setVerbosity(silent)

            self._c_structure = freesasa_structure_from_pdb(input, c, structure_options)

            if not classifier._isCClassifier():
                  setVerbosity(normal)

            fclose(input)

            if self._c_structure is NULL:
                  raise Exception("Error reading '%s'." % fileName)

            # for pure Python classifiers we use the default
            # classifier above to initialize the structure and then
            # reassign radii using the provided classifier here
            if (not classifier._isCClassifier()):
                  self.setRadiiWithClassifier(classifier)


      ## Add atom to structure.
      #
      # This function is meant to be used if the structure was not
      # initialized from a PDB. Default radii will be assigned to each
      # atom. This can be overriden by calling
      # setRadiiWithClassifier() afterwards.
      #
      # There are no restraints on string lengths for the arguments, but
      # the atom won't be added if the default classifier doesn't
      # recognize the atom and also cannot deduce its element from the
      # atom name.
      #      
      # @param atomName (str) atom name (e.g. `"CA"`)
      # @param residueName (str) residue name (e.g. `"ALA"`)
      # @param residueNumber (str or int) residue number (e.g. `'12'`)
      #      or integer. Some PDBs have residue-numbers that aren't 
      #      regular numbers. Therefore treated as a string primarily.
      # @param chainLabel (str) 1-character string with chain label (e.g. 'A')
      # @param x,y,z (float) coordinates
      # 
      # @exception Exception Residue-number invalid
      def addAtom(self, atomName, residueName, residueNumber, chainLabel, x, y, z):
            if (type(residueNumber) is str):
                  resnum = residueNumber
            elif (type(residueNumber) is int):
                  resnum = "%d" % residueNumber
            else:
                  raise Exception("Residue-number invalid, must be either string or number")
            cdef const char *label = chainLabel
            ret = freesasa_structure_add_atom(self._c_structure, atomName,
                                              residueName, resnum, label[0],
                                              x, y, z)
            assert(ret != FREESASA_FAIL)

      ## Assign radii to atoms in structure using a classifier.
      #
      # @param classifier A classifier to use to calculate radii
      # @exception AssertionError if structure not properly initialized
      def setRadiiWithClassifier(self,classifier):
            assert(self._c_structure is not NULL)
            n = self.nAtoms()
            r = []
            for i in range(0,n):
                  r.append(classifier.radius(self.residueName(i), self.atomName(i)))
            self.setRadii(r)

      ## Set atomic radii from an array
      # @param radiusArray Array of atomic radii in Ångström, should 
      #                    have nAtoms() elements.
      # @exception AssertionError if radiusArray has wrong dimension, structure 
      #                           not properly initialized, or if the array contains
      #                           negative radii (not properly classified?) 
      def setRadii(self,radiusArray):
            assert(self._c_structure is not NULL)
            n = self.nAtoms()
            assert len(radiusArray) == n
            cdef double *r = <double *>malloc(sizeof(double)*n)
            assert(r is not NULL)
            for i in range(0,n):
                  r[i] = radiusArray[i]
                  assert(r[i] >= 0)
            freesasa_structure_set_radius(self._c_structure, r)

      ## Number of atoms.
      #
      # @return  Number of atoms
      # @exception AssertionError if not properly initialized
      def nAtoms(self):
            assert(self._c_structure is not NULL)
            return freesasa_structure_n(self._c_structure)

      ## Radius of atom.
      # @param i (int) Index of atom.
      # @return Radius in Å. 
      # @exception AssertionError if index out of bounds, object not properly initalized.
      def radius(self,i):
            assert(i >= 0 and i < self.nAtoms())
            assert(self._c_structure is not NULL)
            cdef const double *r = freesasa_structure_radius(self._c_structure)
            assert(r is not NULL)
            return r[i]

      ## Set radius for a given atom
      # @param atomIndex Index of atom
      # @param radius Value of radius
      # @exception AssertionError if index out of bounds, radius
      #            negative, or structure not properly initialized
      def setRadius(self, atomIndex, radius):
            assert(self._c_structure is not NULL)
            assert(atomIndex >= 0 and atomIndex < self.nAtoms())
            assert(radius >= 0)
            freesasa_structure_atom_set_radius(self._c_structure, atomIndex, radius);
      
      ## Get atom name
      # @param i (int) Atom index.
      # @return Atom name as 4-character string.
      # @exception AssertionError: if index out of range or Structure not properly initialized.
      def atomName(self,i):
            assert(i >= 0 and i < self.nAtoms())
            assert(self._c_structure is not NULL)
            return freesasa_structure_atom_name(self._c_structure,i);
      
      ## Get residue name of given atom.
      # @param i (int) Atom index.
      # @return Residue name as 3-character string.
      # @exception AssertionError if index out of range or Structure not properly initialized
      def residueName(self,i):
            assert(i >= 0 and i < self.nAtoms())
            assert(self._c_structure is not NULL)
            return freesasa_structure_atom_res_name(self._c_structure,i)

      ## Get residue number for given atom
      # @param i (int) Atom index.
      # @return Residue number as 4-character string
      # @exception AssertionError if index out of range or Structure not properly initialized
      def residueNumber(self,i):
            assert(i >= 0 and i < self.nAtoms())
            assert(self._c_structure is not NULL)
            return freesasa_structure_atom_res_number(self._c_structure,i);

      ## Get chain label for given atom.
      # @param i (int) Atom index.
      # @return Chain label as 1-character string
      # @exception AssertionError if index out of range or Structure not properly initialized
      def chainLabel(self,i):
            assert(i >= 0 and i < self.nAtoms())
            assert(self._c_structure is not NULL)
            cdef char label[2]
            label[0] = freesasa_structure_atom_chain(self._c_structure,i); 
            label[1] = '\0'
            return label

      ## Get coordinates of given atom.
      # @param i (int) Atom index.
      # @return array of x, y, and z coordinates
      # @exception AssertionError if index out of range or Structure not properly initialized
      def coord(self, i):
            assert(i >= 0 and i < self.nAtoms())
            assert(self._c_structure is not NULL)
            cdef const double *coord = freesasa_structure_coord_array(self._c_structure);
            return [coord[3*i], coord[3*i+1], coord[3*i+2]]

      @staticmethod
      def _get_structure_options(param):
            options = 0
            
            # check validity of options
            knownOptions = {'hetatm','hydrogen','join-models','separate-models',
                            'separate-chains','skip-unknown','halt-at-unknown'}
            unknownOptions = []
            for key in param:
                  if not key in knownOptions:
                        unknownOptions.append(key)
            if len(unknownOptions) > 0:
                  raise AssertionError("Option(s): ",unknownOptions," unknown.");

            # calculate bitfield
            if 'hetatm' in param and param['hetatm']:
                  options |= FREESASA_INCLUDE_HETATM
            if 'hydrogen' in param and param['hydrogen']:
                  options |= FREESASA_INCLUDE_HYDROGEN
            if 'join-models' in param and param['join-models']:
                  options |= FREESASA_JOIN_MODELS
            if 'separate-models' in param and param['separate-models']:
                  options |= FREESASA_SEPARATE_MODELS
            if 'separate-chains' in param and param['separate-chains']:
                  options |= FREESASA_SEPARATE_CHAINS
            if 'skip-unknown' in param and param['skip-unknown']:
                  options |= FREESASA_SKIP_UNKNOWN
            if 'halt-at-unknown' in param and param['halt-at-unknown']:
                  options |= FREESASA_HALT_AT_UNKNOWN
            return options

      def _get_address(self, size_t ptr2ptr):
            cdef freesasa_structure **p = <freesasa_structure**> ptr2ptr
            p[0] = self._c_structure

      def _set_address(self, size_t ptr2ptr):
            cdef freesasa_structure **p = <freesasa_structure**> ptr2ptr
            self._c_structure = p[0]

      ## The destructor
      def __dealloc__(self):
            if self._c_structure is not NULL:
                  freesasa_structure_free(self._c_structure)

## Default options for structureArray.
# Defined separately for Doxygen's sake.
defaultStructureArrayOptions = {
      'hetatm' : False, 
      'hydrogen' : False,
      'separate-chains' : True,
      'separate-models' : False
}

## Create array of structures from PDB file.
#
# Split PDB file into several structures by either by treating
# chains separately, by treating each MODEL as a separate
# structure, or both.
# 
# @param fileName (str) The PDB file.
# @param options (dict) Specification for how to read the PDB-file
#  (see def value for options).
# @param classifier: Classifier to assign atoms radii, default is used
#   if none specified.
# @exception AssertionError if fileName is None
# @exception AssertionError if an option value is not recognized
# @exception AssertionError if neither of the options 'separate-chains'
#   and 'separate-models' are specified.
# @exception IOError if can't open file
# @exception Exception: if there are problems parsing the input
# @return An array of Structures
def structureArray(fileName,
                   options = defaultStructureArrayOptions,
                   classifier = None):
      assert fileName is not None
      # we need to have at least one of these
      assert(('separate-chains' in options and options['separate-chains'] is True) 
             or ('separate-models' in options and options['separate-models'] is True))
      structure_options = Structure._get_structure_options(options)
      cdef FILE *input
      input = fopen(fileName,'r')
      if input is NULL:
            raise IOError("File '%s' could not be opened." % fileName)
      cdef int n;
      cdef freesasa_structure** sArray = freesasa_structure_array(input,&n,NULL,structure_options)
      fclose(input)
      if sArray is NULL:
            raise Exception("Problems reading structures in '%s'." % fileName)
      structures = []
      for i in range(0,n):
            structures.append(Structure())
            structures[-1]._set_address(<size_t> &sArray[i])
            if classifier is not None:
                  structures[-1].setRadiiWithClassifier(classifier)
      free(sArray)
      return structures


## Calculate SASA of Structure
# @param structure Structure to be used
# @param parameters Parameters to use (if not specified defaults are used)
# @return A Result object
# @exception Exception: something went wrong in calculation (see C library error messages)
def calc(structure,parameters=None):
      cdef const freesasa_parameters *p = NULL
      cdef const freesasa_structure *s = NULL
      if parameters is not None:  parameters._get_address(<size_t>&p)
      structure._get_address(<size_t>&s)
      result = Result()
      result._c_result = <freesasa_result*> freesasa_calc_structure(s,p)
      if result._c_result is NULL:
            raise Exception("Error calculating SASA.")
      return result

## Calculate SASA for a set of coordinates and radii
# @param coord list of size 3*N with atomic coordinates (x1, y1, z1,
#   x2, y2, z2, ..., x_N, y_N, z_N'.
# @param radii array of size N with atomic radii (r_1, r_2, ..., r_N)
# @param Parameters to use (if not specified, defaults are used)
# @exception AssertionError: mismatched array-sizes
# @exception Exception: Out of memory
# @exception Exception: something went wrong in calculation (see C library error messages)
def calcCoord(coord, radii, parameters=None):
      assert(len(coord) == 3*len(radii))

      cdef const freesasa_parameters *p = NULL
      cdef double *c = <double*> malloc(len(coord)*sizeof(double))
      cdef double *r = <double*> malloc(len(radii)*sizeof(double))
      if c is NULL or r is NULL:
            raise Exception("Memory allocation error")

      for i in xrange(len(coord)):
            c[i] = coord[i]
      for i in xrange(len(radii)):
            r[i] = radii[i]

      if parameters is not None: parameters._get_address(<size_t>&p)

      result = Result()
      result._c_result = <freesasa_result*> freesasa_calc_coord(c, r, len(radii), p)

      if result._c_result is NULL:
            raise Exception("Error calculating SASA.")

      free(c)
      free(r)

      return result

## Break SASA result down into classes.
# @param result Result from sasa calculation.
# @param structure Structure used in calculation.
# @param classifier Classifier (if not specified default is used).
# @return Dictionary with names of classes as keys and their SASA values as values.
# @exception Exception: Problems with classification, see C library error messages 
#  (or Python exceptions if run with derived classifier).
def classifyResults(result,structure,classifier=None):
      if classifier is None:
            classifier = Classifier()
      ret = dict()
      for i in range(0,structure.nAtoms()):
            name = classifier.classify(structure.residueName(i),structure.atomName(i))
            if name not in ret:
                  ret[name] = 0
            ret[name] += result.atomArea(i)
      return ret

## Sum SASA result over a selection of atoms
# @param commands A list of commands with selections using Pymol
#   syntax, e.g. `"s1, resn ala+arg"` or `"s2, chain A and resi 1-5"` 
#   (see @ref Selection).
# @param structure A Structure.  
# @param result Result from sasa calculation on structure.
# @return Dictionary with names of selections (`"s1"`,`"s2"`,...) as 
#   keys, and the corresponding SASA values as values.
# @exception Exception: Parser failed (typically syntax error), see
#   library error messages.
def selectArea(commands, structure, result):
      cdef freesasa_structure *s
      cdef freesasa_result *r
      cdef freesasa_selection *selection
      structure._get_address(<size_t> &s)
      result._get_address(<size_t> &r)
      value = dict()
      for cmd in commands:
            selection = freesasa_selection_new(cmd, s, r)
            if selection == NULL:
                  raise Exception("Error parsing '%s'" % cmd)
            value[freesasa_selection_name(selection)] = freesasa_selection_area(selection)
            freesasa_selection_free(selection)
      return value

## Set global verbosity
# @param verbosity Can have values freesasa.silent, freesasa.nowarnings or freesasa.normal
# @exception AssertionError if verbosity has illegal value
def setVerbosity(verbosity):
      assert(verbosity in [silent, nowarnings, normal])
      freesasa_set_verbosity(verbosity)


## Get global verbosity
# @return Verbosity (freesasa.silent, freesasa.nowarnings or freesasa.normal)
def getVerbosity():
      return freesasa_get_verbosity()

## Create a freesasa structure from a Bio.PDB structure
#
#  @remark Experimental, not thorougly tested yet
#
#  @param bioPDBStructure a Bio.PDB structure
#  @param classifier an optional classifier to specify atomic radii
#  @param options Options supported are 'hetatm', 'skip-unknown' and 'halt-at-unknown'
#  @return a freesasa.Structure
#
#  @exception Exception if option 'halt-at-unknown' is selected and
#             unknown atoms are encountered. Passes on exceptions from
#             Structure.addAtom() and
#             Structure.setRadiiWithClassifier().
def structureFromBioPDB(bioPDBStructure, classifier=None, options = Structure.defaultOptions):
      structure = Structure()
      if (classifier is None):
            classifier = Classifier()
      optbitfield = Structure._get_structure_options(options)

      atoms = bioPDBStructure.get_atoms()

      for a in atoms:
            r = a.get_parent()
            hetflag, resseq, icode = r.get_id()

            if (hetflag is not ' ' and not (optbitfield & FREESASA_INCLUDE_HETATM)):
                  continue

            c = r.get_parent()
            v = a.get_vector()

            if (classifier.classify(r.get_resname(), a.get_fullname()) is 'Unknown'):
                  if (optbitfield & FREESASA_SKIP_UNKNOWN):
                        continue
                  if (optbitfield & FREESASA_HALT_AT_UNKNOWN):
                        raise Exception("Halting at unknown atom")

            structure.addAtom(a.get_fullname(), r.get_resname(), resseq, c.get_id(),
                              v[0], v[1], v[2])

      structure.setRadiiWithClassifier(classifier)
      return structure

## Calc SASA from Bio.PDB structure
#
#  Usage 
#
#      result, sasa_classes = calcBioPDB(structure, ...)  
#
#  @remark Experimental, not thorougly tested yet
#
#  @param bioPDBStructure A Bio.PDB structure
#  @param parameters A freesasa.Paremeters object
#  @param classifier A freesasa.Classifier object
#  @param options Options supported are 'hetatm', 'skip-unknown' and 'halt-at-unknown'
#
#  @return A freesasa.Result object and a dictionary with classes
#         defined by the classifier and associated areas
#
#  @exception Exception if unknown atom is encountered and the option
#             'halt-at-unknown' is active. Passes on exceptions from
#             calc(), classifyResults() and structureFromBioPDB().
def calcBioPDB(bioPDBStructure, parameters = Parameters(), 
               classifier = None, options = Structure.defaultOptions):
      structure = structureFromBioPDB(bioPDBStructure, classifier, options)
      result = calc(structure, parameters)
      sasa_classes = classifyResults(result, structure, classifier)
      return result, sasa_classes
      




# -*- coding: utf-8 -*-

import re
import urllib2
import gzip
import os
from StringIO import StringIO
from utils import query_db

from wtforms.validators import DataRequired, Length, Email, optional, \
    ValidationError, NumberRange

from parsePDB import PdbParser
# from config_manager import config

###
### Validators shared by CABS and aggrescan
###


def pdb_input_validator(form, field):
    if form.input_pdb_code.data and form.input_file.data:
        raise ValidationError('Please provide either a PDB code or file, not both.')
    if not form.input_pdb_code.data and not form.input_file.data:
        raise ValidationError('Protein PDB code or PDB file is required')
    if len(field.data) < 4 and not form.input_file.data:
        raise ValidationError('Protein code must be 4-letter (2PCY).'
                              'Leave empty only if PDB file is provided')


def sequence_validator(form, field):
    allowed_seq = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N',
                   'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']
    for letter in re.sub("\s", "", field.data):
        if letter not in allowed_seq:
            raise ValidationError('Sequence contains non-standard aminoacid \
                                  symbol: %s' % letter)


def pdb_file_validator(form, field):
    if form.input_file.data:
        p = PdbParser(form.input_file.data.stream,  form.chain.data.strip())
        _structure_pdb_validator(p)


def pdb_code_validator(form, field):
    if len(form.input_pdb_code.data.strip()) == 4:
        pdb_code = field.data.strip()
        try:
            buraki = urllib2.urlopen('http://www.rcsb.org/pdb/files/'+pdb_code+'.pdb.gz')
        except:
            raise ValidationError("Could not download your protein file from pdb server. Wrong code perhaps?")

        b2 = buraki.read()
        ft = StringIO(b2)
        buraki.close()
        with gzip.GzipFile(fileobj=ft, mode="rb") as f:
            p = PdbParser(f)
            _structure_pdb_validator(p, form.chain.data.strip())

    elif not form.input_pdb_code.data:
        pass
    else:
        raise ValidationError("PDB code should contain exactly 4 characters.")


def non_empty_file(form, field):
    """To be used with files only"""
    if field.data:
        field.data.stream.seek(0, os.SEEK_END)
        if field.data.stream.tell() < 1:
            raise ValidationError("Empty file provided.")
        field.data.stream.seek(0)


def _structure_pdb_validator(p, chains=''):
    """
    :param p: parsePDB PdbParser object
    :param chains: a list of chain (might also be an empty string)
    :return: None, raises ValidationError if any errors are found
    """
    missing = p.getMissing()
    seq = p.getSequence()
    n_chains = p.get_num_chains()
    defined_chains = set(chains)
    actual_chains = set(p.getChains())
    allowed_seq = ['A', 'C', 'D', ' ', 'E', 'F', 'G', 'H', 'I', 'K', 'L',
                   'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']
    _max_length = 2000
    _max_chains = 10
    for e in seq:
        if e not in allowed_seq:
            raise ValidationError('Non-standard residue in the input \
                    structure')
    for length in p.getChainLengths():
        if length < 4:
            raise ValidationError("One of the chains in the supplied pdb file is too short."
                                  " Minimal chain length for the simulation is 4.")
    if len(seq) == 0:
        raise ValidationError('Your pdb file doesnt seem to contain a chain. The server could not extract a sequence.')
    if len(seq) > _max_length:
        raise ValidationError('CABS-flex allows max %d receptor \
                               residues. Provided file contains \
                               %d' % (_max_length, len(seq)))
    if n_chains > _max_chains:
        raise ValidationError('CABS-flex allows a maximum of %d chains.' 
                              'Your protein has: %d' % (n_chains, _max_chains))
    if missing > 5:
        raise ValidationError('Missing atoms within protein (M+N = %d). \
                Protein must fulfill M+N<6, where M - number of \
                chains, N - number of breaks' % (missing))
    if not defined_chains.issubset(actual_chains):
        raise ValidationError("Wrong chain selection. Your PDB contains chain(s): %s and you tried to select chain(s): %s" %
                              (", ".join(actual_chains), ", ".join(defined_chains)))


def only_letters_validator(form, field):
    pattern = re.compile("[^a-zA-Z\s]")
    if pattern.findall(field.data):
        raise ValidationError("Only letters and whitespaces are allowed in the chain field. You entered: %s" % field.data)

###
### Aggrescan specific validators
###


def foldx_validator(form, field):
    """check if a valid foldx loc is stored in the database"""
    if field.data == '1':
        foldx_path = query_db("SELECT foldx_path FROM system_data WHERE id=1", one=True)[0]
        req_file = os.path.join(foldx_path, 'rotabase.txt')
        if not os.path.isfile(req_file):
            raise ValidationError("Trying to run Stability calculations without a proper FoldX path specified."
                                  " (%s missing rotabase.txt file)" % foldx_path)


def foldx_and_m_validator(form, field):
    if field.data == "1" and form.foldx.data != "1":
        raise ValidationError("Please select the Stability calculations option if you wish to create a mutant.")


def single_chain_validator(form, field):
    pattern = re.compile("[^a-zA-Z\s]")
    if pattern.findall(field.data):
        raise ValidationError("Only letters and whitespaces are allowed in the chain field. You entered: %s" % field.data)
    elif len(field.data.replace(" ", "")) > 1:
        raise ValidationError("Please select a single chain")


def auto_mut_validator(form, field):
    if field.data == "1" and form.foldx.data != "1":
        raise ValidationError("Please select the Stability calculations option if you wish to use mutation options.")
    if field.data == "1" and form.dynamic.data == "1":
        raise ValidationError("Automatic mutations do not support the dynamic module. Please select either of those")


# -*- coding: utf-8 -*-

import os
import json
import re
import sqlite3
import numpy as np
from os.path import join
from glob import glob
from utils import query_db, status_color, sort_dict
from a3d_gui import app
from collections import OrderedDict
from aggrescan.optparser import parse_config_file, get_parser_object, _parse_mut


def generate_config(jid):
    """
    Function that that reads a database and creates a config file
    that then can be easily run by Aggrescan3D on the cluster
    paths to cabs and foldx might need to be changed if changes happen
    :param jid:  job unique id
    :return: None
    """

    # There should be exactly one entry in that table. The location needs to be validated on submit
    foldx_loc = query_db("SELECT foldx_path FROM system_data WHERE id=1", one=True)[0]
    file_header = "# Generated for job: %s" % jid
    try:
        project_settings = query_db("SELECT dynamic, mutate, mutation, chains, \
                   foldx, distance, auto_mutation FROM project_details WHERE jid=?", [jid], one=True)
    except sqlite3.OperationalError:  # Backwards compatibility when the auto_mutation column didn't exist
        query_db("ALTER TABLE project_details ADD COLUMN auto_mutation TEXT", insert=True)
        project_settings = query_db("SELECT dynamic, mutate, mutation, chains, \
                           foldx, distance, auto_mutation FROM project_details WHERE jid=?", [jid], one=True)

    project_name = query_db("SELECT project_name FROM user_queue WHERE jid=?", [jid], one=True)
    with open(join(app.config['USERJOB_DIRECTORY'], jid, "config.ini"), 'w') as f:
        _write(f, file_header)
        _write(f, "# The job can also be identified by its name: %s" % project_name['project_name'])
        _write(f, "v = 3")
        _write(f, "protein = input.pdb")
        _write(f, "movie = webm")
        _write(f, "distance = %s" % project_settings["distance"])
        _write(f, "remote")
        if project_settings["chains"]:
            if len(project_settings["chains"]) == 1:    # Aggrescan only accepts 1 specific chain
                _write(f, "chain = %s" % project_settings["chains"])
        if project_settings["foldx"]:
            _write(f, "foldx = %s" % foldx_loc)
        if project_settings["dynamic"]:
            _write(f, "dynamic")
        if project_settings["auto_mutation"]:
            _write(f, "auto_mutation = %s" % project_settings["auto_mutation"])
        if project_settings["mutation"]:
            for entry in project_settings["mutation"].split():
                _write(f, "m = %s" % entry)


def reverse_config(filepath):
    argv, mutations = parse_config_file(filepath)
    parser = get_parser_object()
    options = parser.parse_args(argv)
    mutations = _parse_mut(mutations)
    final_options = {
        'dynamic': options.dynamic,
        'distance': options.distance,
        'chain': options.chain,
        "foldx": options.foldx,
        'auto_mutation': options.auto_mutation,
    }
    print final_options, mutations
    return final_options, mutations


def prepare_data(jid):
    """
    Pull all the necessary data from the database and prepare an OrderedDict
    The dict is then used to display the job_info page
    raises an IO error if there is no data in current database for the specified job ID
    :param jid: unique job ID
    :return: dict of option:value pairs
    """
    system_info = query_db("SELECT chain_sequence, distance, mutt_energy_diff, dynamic, \
            mutation, mutate, chains, foldx, auto_mutation FROM  project_details WHERE jid=?", [jid], one=True)
    basic_info = query_db("SELECT started, project_name, status , working_dir FROM  "
                          "user_queue WHERE jid=?", [jid], one=True)
    project_info = OrderedDict()
    if not basic_info or not system_info:
        raise IOError
    project_info['status'] = basic_info['status']
    project_info['status_color'] = status_color(project_info['status'])
    # Example projects will not exist by their abspath on users machine
    if os.path.exists(os.path.join(app.config['USERJOB_DIRECTORY'], basic_info['working_dir'].split("/")[-1])):
        project_info['working_dir'] = os.path.join(app.config['USERJOB_DIRECTORY'], basic_info['working_dir'].split("/")[-1])
    else:
        project_info['working_dir'] = basic_info['working_dir']
    # csv table reading
    mut = {}
    if system_info['mutate'] == 1:
        for mutation in system_info['mutation'].strip().replace(" ", "").split(","):
            k = mutation[-1] + mutation[2:-1]  # the last letter is the chain ID letter, starting from index 2 is the cahin ID number
            mut[k] = mutation
    if system_info['foldx']:  # foldx is app
        project_info['foldx'] = "Yes"
    else:
        project_info['foldx'] = "No"

    avg_scores = ''
    a3d_table = []
    a3dtable = ''
    chains = set()
    models = ["input.pdb"]
    models.extend(["model_%s.pdb" % str(i) for i in range(12)])
    project_info['models'] = models
    project_info['error'] = ''
    project_info['table'] = ''
    project_info['avg_scores'] = {'dummy': 'dummy'}     # click highest model would cause template errors otheriwse for static jobs
    project_info['chains'] = system_info['chains'] or ""
    project_info['distance'] = system_info['distance']
    project_info['chain_sequence'] = system_info['chain_sequence']
    project_info['project_name'] = basic_info['project_name']
    project_info['started'] = basic_info['started']
    project_info['mutation'] = mut
    project_info['mutate'] = system_info['mutate']
    project_info['mutt_energy_diff'] = system_info['mutt_energy_diff']
    project_info['auto_mutation'] = False
    project_info['auto_mutation_used'] = system_info['auto_mutation']
    project_info['autom_data'] = {'dummy': 'dummy'}
    if system_info['dynamic'] == 1:
        project_info['dynamic'] = True
        pdb_in_dir = set([os.path.basename(i) for i in glob(join(project_info['working_dir'], "*.pdb"))])
        models = set(models)
        if len(models-pdb_in_dir) > 0:
            project_info['model_files'] = 'missing'
        else:
            project_info['model_files'] = 'ok'
    else:
        project_info['dynamic'] = False
    if project_info['status'] == 'done':
        if system_info['auto_mutation']:
            project_info['autom_data'] = _parse_auto_mut_info(
                join(project_info['working_dir'], 'Mutations_summary.csv'))
            project_info['auto_mutation'] = True
        if project_info['dynamic']:
            with open(os.path.join(project_info['working_dir'], "averages")) as f:
                loaded_data = json.load(f)
            project_info['avg_scores'] = sort_dict(loaded_data)

        with open(os.path.join(project_info['working_dir'],
                  "A3D.csv")) as fw:
            rec = re.compile(r"^(.*),(.*),(.*),(.*),(.*)$", re.M)
            d = rec.findall(fw.read().replace("\r", ""))[1:]
            dat = []
            for row in d:
                if len(row) != 5:
                    continue
                chain = row[1]
                chains.update(chain)
                residx = row[2]
                resname = row[3]
                v = float(row[4])
                dat.append(v)
                a3d_table.append((residx, resname, chain, "%01.4f" % (v)))
            min3d = min(dat)
            max3d = max(dat)
            sum3d = np.sum(dat)
            avg3d = sum3d/len(dat)
            a3dtable = {'min': min3d, 'avg': avg3d, 'max': max3d,
                        'sum': sum3d, 'tab': a3d_table}
            project_info['table'] = a3dtable
    return project_info


def _write(obj, text):
    obj.write(text)
    obj.write("\n")


def _parse_mut(mut_list):
    """Necessary for now as those come as a list of dicts and the db needs a simple string"""
    return ["%s%s%s%s" % (i['oldres'], i['newres'], i['idx'], i['chain']) for i in mut_list]


def _parse_auto_mut_info(filename):
    """
    Reads the file and returns an ordered dict with the first column as keys, and the next 3 as a list
    Currently will  return top x results - another parameter to decide on
    """
    data = OrderedDict()
    max_results = 12
    counter = 0
    try:
        with open(filename, 'r') as f:
            f.readline()    # Skip first line with labels
            for line in f:
                counter += 1
                parsed = line.split(",")
                data[parsed[0]] = [float(parsed[1]), float(parsed[2]), float(parsed[3])]
                if counter >= max_results:
                    break
    except IOError:
        return {'Data missing': False}  # The template will recognize this as a sign not to load the tab
    if not data:    # File is empty, the likely case is that there were no suitable mutations found
        data = {'No mutants': False}
    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import gzip
from re import compile, sub
import json


def arraytostring(ar):
    i = ["%d-%d" % e for e in ar]
    return ", ".join(i)


class PdbParser:
    def __init__(self, filehandler, chain=''):
        self.codification = {"ALA" : 'A', "CYS" : 'C', "ASP" : 'D', "GLU" : 'E', "PHE" : 'F', "GLY" : 'G', "HIS" : 'H', "ILE" : 'I', "LYS" : 'K', "LEU" : 'L', "MET" : 'M', "MSE" : 'M', "ASN" : 'N', "PYL" : 'O', "PRO" : 'P', "GLN" : 'Q', "ARG" : 'R', "SER" : 'S', "THR" : 'T', "SEC" : 'U', "VAL" : 'V', "TRP" : 'W', "5HP" : 'E', "ABA" : 'A', "AIB" : 'A', "BMT" : 'T', "CEA" : 'C', "CGU" : 'E', "CME" : 'C', "CRO" : 'X', "CSD" : 'C', "CSO" : 'C', "CSS" : 'C', "CSW" : 'C', "CSX" : 'C', "CXM" : 'M', "DAL" : 'A', "DAR" : 'R', "DCY" : 'C', "DGL" : 'E', "DGN" : 'Q', "DHI" : 'H', "DIL" : 'I', "DIV" : 'V', "DLE" : 'L', "DLY" : 'K', "DPN" : 'F', "DPR" : 'P', "DSG" : 'N', "DSN" : 'S', "DSP" : 'D', "DTH" : 'T', "DTR" : 'X', "DTY" : 'Y', "DVA" : 'V', "FME" : 'M', "HYP" : 'P', "KCX" : 'K', "LLP" : 'K', "MLE" : 'L', "MVA" : 'V', "NLE" : 'L', "OCS" : 'C', "ORN" : 'A', "PCA" : 'E', "PTR" : 'Y', "SAR" : 'G', "SEP" : 'S', "STY" : 'Y', "TPO" : 'T', "TPQ" : 'F', "TYS" : 'Y', "TYR" : 'Y' }
        keys = self.codification.keys()
        self.sequences = {}
        self.onlycalfa = ""
        self.chain = chain
        self.canumber = 0
        self.allnumber = 0
        seq = compile(r"^ATOM.{9}CA..(?P<seqid>.{3}).(?P<chain>.{1})(?P<resid>.{4})")  # TODO zle dla alternatywnych
        if chain != '':
            ch = "|".join(list(chain))
            seq_c = compile(r"^ATOM.{9}CA.( |A)(?P<seqid>.{3}).("+ch+")")
            atm =   compile(r"^ATOM.{9}(.{3})( |A)(?P<seqid>.{3}).("+ch+")(?P<resid>.{4})(?P<x>.{12})(?P<y>.{8})(?P<z>.{8})")
        else:
            atm =   compile(r"^ATOM.{9}(.{3})( |A)(?P<seqid>.{3})..(?P<resid>.{4})(?P<x>.{12})(?P<y>.{8})(?P<z>.{8})")
            seq_c = compile(r"^ATOM.{9}CA.( |A)(?P<seqid>.{3})")

        ter = compile(r"^TER")  # TODO nie jestem pewien
        mod = compile(r"^ENDMDL")  # TODO nie jestem pewien
        self.trajectory = []
        self.sequence = ""
        self.resindexes = []
        self.mutatedata = {}
        f = filehandler
        lines = f.readlines()
        end = len(lines)-1
        counter = 0
        tmp = []
        chains_order = []
        old_chain_id = ''
        chain_break = False     # a band-aid to fix crashes when TER records are missing
        for line in lines:
            line = self._replaceExotic(line)
            #line = sub(r'^HETATM(.{10}(A| ))MSE', r'ATOM  \1MET',line)
            data = atm.match(line)
            data_seq = seq.match(line)
            data_seqc = seq_c.match(line)

            if data_seqc and data_seqc.group('seqid') in keys:
                self.canumber += 1
            if data_seq:
                seqid = data_seq.group('seqid').strip()
                if seqid not in self.codification.keys():
                    continue
                chainid = data_seq.group('chain').strip()
                if chainid != old_chain_id:
                    if old_chain_id != '':
                        chain_break = True
                    old_chain_id = chainid

                if chainid not in chains_order:
                    chains_order.append(chainid)
                resid = data_seq.group('resid').strip()

                if seqid in keys:
                    self.sequence += self.codification[seqid]
                    s = self.codification[seqid]
                else:
                    s = "X"
                    self.sequence += "X"

                if chainid in self.sequences.keys():
                    self.sequences[chainid] += s
                else:
                    self.sequences[chainid] = s

                if chainid in self.mutatedata.keys():
                    self.mutatedata[chainid].append({'chain': chainid,
                                                     'resname': s,
                                                     'residx': resid})
                else:
                    self.mutatedata[chainid] = [{'chain': chainid,
                                                 'resname': s,
                                                 'residx': resid}]

                if chainid in self.sequences.keys():
                    self.sequences[chainid] += s
                else:
                    self.sequences[chainid] = s
            if data:
                seqid = data.group('seqid').strip()
                if seqid not in self.codification.keys():
                    continue
                self.allnumber += 1
                self.onlycalfa += line
                dg = data.groups()
                if dg[0].strip() == 'CA':
                    try:
                        tmp.append(int(data.group('resid')))
                    except ValueError:
                        raise ValueError("Wrong residue index number. "
                                         "We do not support negative residue indexes which is the most likely cause for this error."
                                         " Offending line : %s" % line)
            if (ter.match(line) or counter == end) or chain_break:
                self.resindexes.append(tmp)
                tmp = []
                chain_break = False
                #self.onlycalfa += line # skip TER in output
            if (mod.match(line) and len(self.onlycalfa) > 1) or counter == end:
                break
            counter += 1
        filehandler.seek(0)

# chain numbering
        o = {}
        for i in range(len(chains_order)):
            o[chains_order[i]] = self.resindexes[i]
        self.numb = o

    def _replaceExotic(self, line):
        if line.startswith("HETATM") and len(line)>21:
            res = line[17:20]
            if res in self.nonstandard:
                sf = r'^HETATM(.{10}(A| ))'+res
                st = r'ATOM  \1'+self.modres[res]
                line = sub(sf, st, line)
        return line

    def isSingleChain(self):
        if self.chain != '' or len(self.sequences.keys()) == 1:
            return True
        else:
            return False

    def containsOnlyCA(self):
        if self.allnumber == self.canumber:
            return True
        else:
            return False

    def getMissing(self):
        return self.isBroken()+len(self.sequences.keys())

    def isBroken(self):
        brk = []
        for j in self.numb.keys():
            indexes = self.numb[j]
            if len(indexes) == 0: # empty chains
                continue
            first = indexes[0]
            for i in range(1, len(indexes)):
                if indexes[i] - 1 != first:
                    brk.append(str(first) + "-" + str(indexes[i]))
                first = indexes[i]
        return len(brk)

    def getResIndexes(self):
        return ",".join([str(i) for i in self.numb[self.chain]])

    def getChains(self):
        return self.sequences.keys()

    def getBody(self):
        t = []
# remove alt indices from pdb
        for line in self.onlycalfa.split("\n"):
            t.append(line[:16]+" "+line[17:])
        return "\n".join(t)

    def savePdbFile(self, outfilename):
        with gzip.open(outfilename, "wb") as fw:
            fw.write(self.getBody())

    def containsChain(self, chain):
        if chain in self.sequences.keys():
            return True

    def get_num_chains(self):
        return len(self.sequences.keys())

    def getSequence(self):
        if self.chain != '':
            o = ""
            for e in list(self.chain):
                if e in self.sequences.keys():
                    o += self.sequences[e]
            return o
        else:
            out = ''
            for k in self.sequences.keys():
                out += "".join(self.sequences[k])
            return out

    def getChainIdxResname(self):
        if self.chain == '':
            return json.dumps(self.mutatedata)
        else:
            return json.dumps({self.chain: self.mutatedata[self.chain]})

    def getChainLengths(self):
        out = []
        if self.chain != '':
            for e in list(self.chain):
                if self.containsChain(e):   # This should be a given but not sure right now
                    out.append(len(self.sequences[e]))
        else:
            for k in self.sequences.keys():
                out.append(len(self.sequences[k]))
        return out


    modres={
                    '0CS':'ALA',                    ##  0CS ALA  3-[(S)-HYDROPEROXYSULFINYL]-L-ALANINE
                    '1AB':'PRO',                    ##  1AB PRO  1,4-DIDEOXY-1,4-IMINO-D-ARABINITOL
                    '1LU':'LEU',                    ##  1LU LEU  4-METHYL-PENTANOIC ACID-2-OXYL GROUP
                    '1PA':'PHE',                    ##  1PA PHE  PHENYLMETHYLACETIC ACID ALANINE
                    '1TQ':'TRP',                    ##  1TQ TRP  6-(FORMYLAMINO)-7-HYDROXY-L-TRYPTOPHAN
                    '1TY':'TYR',                    ##  1TY TYR
                    '23F':'PHE',                    ##  23F PHE  (2Z)-2-AMINO-3-PHENYLACRYLIC ACID
                    '23S':'TRP',                    ##  23S TRP  MODIFIED TRYPTOPHAN
                    '2BU':'ALA',                    ##  2BU ADE
                    '2ML':'LEU',                    ##  2ML LEU  2-METHYLLEUCINE
                    '2MR':'ARG',                    ##  2MR ARG  N3, N4-DIMETHYLARGININE
                    '2MT':'PRO',                    ##  2MT PRO
                    '2OP':'ALA',                    ##  2OP (2S  2-HYDROXYPROPANAL
                    '2TY':'TYR',                    ##  2TY TYR
                    '32S':'TRP',                    ##  32S TRP  MODIFIED TRYPTOPHAN
                    '32T':'TRP',                    ##  32T TRP  MODIFIED TRYPTOPHAN
                    '3AH':'HIS',                    ##  3AH HIS
                    '3MD':'ASP',                    ##  3MD ASP  2S,3S-3-METHYLASPARTIC ACID
                    '3TY':'TYR',                    ##  3TY TYR  MODIFIED TYROSINE
                    '4DP':'TRP',                    ##  4DP TRP
                    '4F3':'ALA',                    ##  4F3 ALA  CYCLIZED
                    '4FB':'PRO',                    ##  4FB PRO  (4S)-4-FLUORO-L-PROLINE
                    '4FW':'TRP',                    ##  4FW TRP  4-FLUOROTRYPTOPHANE
                    '4HT':'TRP',                    ##  4HT TRP  4-HYDROXYTRYPTOPHAN
                    '4IN':'TRP',                    ##  4IN TRP  4-AMINO-L-TRYPTOPHAN
                    '4PH':'PHE',                    ##  4PH PHE  4-METHYL-L-PHENYLALANINE
                    '5CS':'CYS',                    ##  5CS CYS
                    '6CL':'LYS',                    ##  6CL LYS  6-CARBOXYLYSINE
                    '6CW':'TRP',                    ##  6CW TRP  6-CHLORO-L-TRYPTOPHAN
                    'A0A':'ASP',                    ##  A0A ASP  ASPARTYL-FORMYL MIXED ANHYDRIDE
                    'AA4':'ALA',                    ##  AA4 ALA  2-AMINO-5-HYDROXYPENTANOIC ACID
                    'AAR':'ARG',                    ##  AAR ARG  ARGININEAMIDE
                    'AB7':'GLU',                    ##  AB7 GLU  ALPHA-AMINOBUTYRIC ACID
                    'ABA':'ALA',                    ##  ABA ALA  ALPHA-AMINOBUTYRIC ACID
                    'ACB':'ASP',                    ##  ACB ASP  3-METHYL-ASPARTIC ACID
                    'ACL':'ARG',                    ##  ACL ARG  DEOXY-CHLOROMETHYL-ARGININE
                    'ACY':'GLY',                    ##  ACY GLY  POST-TRANSLATIONAL MODIFICATION
                    'AEI':'THR',                    ##  AEI THR  ACYLATED THR
                    'AFA':'ASN',                    ##  AFA ASN  N-[7-METHYL-OCT-2,4-DIENOYL]ASPARAGINE
                    'AGM':'ARG',                    ##  AGM ARG  4-METHYL-ARGININE
                    'AGT':'CYS',                    ##  AGT CYS  AGMATINE-CYSTEINE ADDUCT
                    'AHB':'ASN',                    ##  AHB ASN  BETA-HYDROXYASPARAGINE
                    'AHO':'ALA',                    ##  AHO ALA  N-ACETYL-N-HYDROXY-L-ORNITHINE
                    'AHP':'ALA',                    ##  AHP ALA  2-AMINO-HEPTANOIC ACID
                    'AIB':'ALA',                    ##  AIB ALA  ALPHA-AMINOISOBUTYRIC ACID
                    'AKL':'ASP',                    ##  AKL ASP  3-AMINO-5-CHLORO-4-OXOPENTANOIC ACID
                    'ALA':'ALA',                    ##  ALA ALA
                    'ALC':'ALA',                    ##  ALC ALA  2-AMINO-3-CYCLOHEXYL-PROPIONIC ACID
                    'ALG':'ARG',                    ##  ALG ARG  GUANIDINOBUTYRYL GROUP
                    'ALM':'ALA',                    ##  ALM ALA  1-METHYL-ALANINAL
                    'ALN':'ALA',                    ##  ALN ALA  NAPHTHALEN-2-YL-3-ALANINE
                    'ALO':'THR',                    ##  ALO THR  ALLO-THREONINE
                    'ALS':'ALA',                    ##  ALS ALA  2-AMINO-3-OXO-4-SULFO-BUTYRIC ACID
                    'ALT':'ALA',                    ##  ALT ALA  THIOALANINE
                    'ALY':'LYS',                    ##  ALY LYS  N(6)-ACETYLLYSINE
                    'AME':'MET',                    ##  AME MET  ACETYLATED METHIONINE
                    'AP7':'ALA',                    ##  AP7 ADE
                    'APH':'ALA',                    ##  APH ALA  P-AMIDINOPHENYL-3-ALANINE
                    'API':'LYS',                    ##  API LYS  2,6-DIAMINOPIMELIC ACID
                    'APK':'LYS',                    ##  APK LYS
                    'AR2':'ARG',                    ##  AR2 ARG  ARGINYL-BENZOTHIAZOLE-6-CARBOXYLIC ACID
                    'AR4':'GLU',                    ##  AR4 GLU
                    'ARG':'ARG',                    ##  ARG ARG
                    'ARM':'ARG',                    ##  ARM ARG  DEOXY-METHYL-ARGININE
                    'ARO':'ARG',                    ##  ARO ARG  C-GAMMA-HYDROXY ARGININE
                    'ASA':'ASP',                    ##  ASA ASP  ASPARTIC ALDEHYDE
                    'ASB':'ASP',                    ##  ASB ASP  ASPARTIC ACID-4-CARBOXYETHYL ESTER
                    'ASI':'ASP',                    ##  ASI ASP  L-ISO-ASPARTATE
                    'ASK':'ASP',                    ##  ASK ASP  DEHYDROXYMETHYLASPARTIC ACID
                    'ASL':'ASP',                    ##  ASL ASP  ASPARTIC ACID-4-CARBOXYETHYL ESTER
                    'ASN':'ASN',                    ##  ASN ASN
                    'ASP':'ASP',                    ##  ASP ASP
                    'AYA':'ALA',                    ##  AYA ALA  N-ACETYLALANINE
                    'AYG':'ALA',                    ##  AYG ALA
                    'AZK':'LYS',                    ##  AZK LYS  (2S)-2-AMINO-6-TRIAZANYLHEXAN-1-OL
                    'B2A':'ALA',                    ##  B2A ALA  ALANINE BORONIC ACID
                    'B2F':'PHE',                    ##  B2F PHE  PHENYLALANINE BORONIC ACID
                    'B2I':'ILE',                    ##  B2I ILE  ISOLEUCINE BORONIC ACID
                    'B2V':'VAL',                    ##  B2V VAL  VALINE BORONIC ACID
                    'B3A':'ALA',                    ##  B3A ALA  (3S)-3-AMINOBUTANOIC ACID
                    'B3D':'ASP',                    ##  B3D ASP  3-AMINOPENTANEDIOIC ACID
                    'B3E':'GLU',                    ##  B3E GLU  (3S)-3-AMINOHEXANEDIOIC ACID
                    'B3K':'LYS',                    ##  B3K LYS  (3S)-3,7-DIAMINOHEPTANOIC ACID
                    'B3S':'SER',                    ##  B3S SER  (3R)-3-AMINO-4-HYDROXYBUTANOIC ACID
                    'B3X':'ASN',                    ##  B3X ASN  (3S)-3,5-DIAMINO-5-OXOPENTANOIC ACID
                    'B3Y':'TYR',                    ##  B3Y TYR
                    'BAL':'ALA',                    ##  BAL ALA  BETA-ALANINE
                    'BBC':'CYS',                    ##  BBC CYS
                    'BCS':'CYS',                    ##  BCS CYS  BENZYLCYSTEINE
                    'BCX':'CYS',                    ##  BCX CYS  BETA-3-CYSTEINE
                    'BFD':'ASP',                    ##  BFD ASP  ASPARTATE BERYLLIUM FLUORIDE
                    'BG1':'SER',                    ##  BG1 SER
                    'BHD':'ASP',                    ##  BHD ASP  BETA-HYDROXYASPARTIC ACID
                    'BIF':'PHE',                    ##  BIF PHE
                    'BLE':'LEU',                    ##  BLE LEU  LEUCINE BORONIC ACID
                    'BLY':'LYS',                    ##  BLY LYS  LYSINE BORONIC ACID
                    'BMT':'THR',                    ##  BMT THR
                    'BNN':'ALA',                    ##  BNN ALA  ACETYL-P-AMIDINOPHENYLALANINE
                    'BOR':'ARG',                    ##  BOR ARG
                    'BPE':'CYS',                    ##  BPE CYS
                    'BTR':'TRP',                    ##  BTR TRP  6-BROMO-TRYPTOPHAN
                    'BUC':'CYS',                    ##  BUC CYS  S,S-BUTYLTHIOCYSTEINE
                    'BUG':'LEU',                    ##  BUG LEU  TERT-LEUCYL AMINE
                    'C12':'ALA',                    ##  C12 ALA
                    'C1X':'LYS',                    ##  C1X LYS  MODIFIED LYSINE
                    'C3Y':'CYS',                    ##  C3Y CYS  MODIFIED CYSTEINE
                    'C5C':'CYS',                    ##  C5C CYS  S-CYCLOPENTYL THIOCYSTEINE
                    'C6C':'CYS',                    ##  C6C CYS  S-CYCLOHEXYL THIOCYSTEINE
                    'C99':'ALA',                    ##  C99 ALA
                    'CAB':'ALA',                    ##  CAB ALA  4-CARBOXY-4-AMINOBUTANAL
                    'CAF':'CYS',                    ##  CAF CYS  S-DIMETHYLARSINOYL-CYSTEINE
                    'CAS':'CYS',                    ##  CAS CYS  S-(DIMETHYLARSENIC)CYSTEINE
                    'CCS':'CYS',                    ##  CCS CYS  CARBOXYMETHYLATED CYSTEINE
                    'CGU':'GLU',                    ##  CGU GLU  CARBOXYLATION OF THE CG ATOM
                    'CH6':'ALA',                    ##  CH6 ALA
                    'CH7':'ALA',                    ##  CH7 ALA
                    'CHG':'GLY',                    ##  CHG GLY  CYCLOHEXYL GLYCINE
                    'CHP':'GLY',                    ##  CHP GLY  3-CHLORO-4-HYDROXYPHENYLGLYCINE
                    'CHS':'PHE',                    ##  CHS PHE  4-AMINO-5-CYCLOHEXYL-3-HYDROXY-PENTANOIC AC
                    'CIR':'ARG',                    ##  CIR ARG  CITRULLINE
                    'CLB':'ALA',                    ##  CLB ALA
                    'CLD':'ALA',                    ##  CLD ALA
                    'CLE':'LEU',                    ##  CLE LEU  LEUCINE AMIDE
                    'CLG':'LYS',                    ##  CLG LYS
                    'CLH':'LYS',                    ##  CLH LYS
                    'CLV':'ALA',                    ##  CLV ALA
                    'CME':'CYS',                    ##  CME CYS  MODIFIED CYSTEINE
                    'CML':'CYS',                    ##  CML CYS
                    'CMT':'CYS',                    ##  CMT CYS  O-METHYLCYSTEINE
                    'CQR':'ALA',                    ##  CQR ALA
                    'CR2':'ALA',                    ##  CR2 ALA  POST-TRANSLATIONAL MODIFICATION
                    'CR5':'ALA',                    ##  CR5 ALA
                    'CR7':'ALA',                    ##  CR7 ALA
                    'CR8':'ALA',                    ##  CR8 ALA
                    'CRK':'ALA',                    ##  CRK ALA
                    'CRO':'THR',                    ##  CRO THR  CYCLIZED
                    'CRQ':'TYR',                    ##  CRQ TYR
                    'CRW':'ALA',                    ##  CRW ALA
                    'CRX':'ALA',                    ##  CRX ALA
                    'CS1':'CYS',                    ##  CS1 CYS  S-(2-ANILINYL-SULFANYL)-CYSTEINE
                    'CS3':'CYS',                    ##  CS3 CYS
                    'CS4':'CYS',                    ##  CS4 CYS
                    'CSA':'CYS',                    ##  CSA CYS  S-ACETONYLCYSTEIN
                    'CSB':'CYS',                    ##  CSB CYS  CYS BOUND TO LEAD ION
                    'CSD':'CYS',                    ##  CSD CYS  3-SULFINOALANINE
                    'CSE':'CYS',                    ##  CSE CYS  SELENOCYSTEINE
                    'CSI':'ALA',                    ##  CSI ALA
                    'CSO':'CYS',                    ##  CSO CYS  INE S-HYDROXYCYSTEINE
                    'CSR':'CYS',                    ##  CSR CYS  S-ARSONOCYSTEINE
                    'CSS':'CYS',                    ##  CSS CYS  1,3-THIAZOLE-4-CARBOXYLIC ACID
                    'CSU':'CYS',                    ##  CSU CYS  CYSTEINE-S-SULFONIC ACID
                    'CSW':'CYS',                    ##  CSW CYS  CYSTEINE-S-DIOXIDE
                    'CSX':'CYS',                    ##  CSX CYS  OXOCYSTEINE
                    'CSY':'ALA',                    ##  CSY ALA  MODIFIED TYROSINE COMPLEX
                    'CSZ':'CYS',                    ##  CSZ CYS  S-SELANYL CYSTEINE
                    'CTH':'THR',                    ##  CTH THR  4-CHLOROTHREONINE
                    'CWR':'ALA',                    ##  CWR ALA
                    'CXM':'MET',                    ##  CXM MET  N-CARBOXYMETHIONINE
                    'CY0':'CYS',                    ##  CY0 CYS  MODIFIED CYSTEINE
                    'CY1':'CYS',                    ##  CY1 CYS  ACETAMIDOMETHYLCYSTEINE
                    'CY3':'CYS',                    ##  CY3 CYS  2-AMINO-3-MERCAPTO-PROPIONAMIDE
                    'CY4':'CYS',                    ##  CY4 CYS  S-BUTYRYL-CYSTEIN
                    'CY7':'CYS',                    ##  CY7 CYS  MODIFIED CYSTEINE
                    'CYD':'CYS',                    ##  CYD CYS
                    'CYF':'CYS',                    ##  CYF CYS  FLUORESCEIN LABELLED CYS380 (P14)
                    'CYG':'CYS',                    ##  CYG CYS
                    'CYJ':'LYS',                    ##  CYJ LYS  MODIFIED LYSINE
                    'CYQ':'CYS',                    ##  CYQ CYS
                    'CYR':'CYS',                    ##  CYR CYS
                    'CYS':'CYS',                    ##  CYS CYS
                    'CZ2':'CYS',                    ##  CZ2 CYS  S-(DIHYDROXYARSINO)CYSTEINE
                    'CZZ':'CYS',                    ##  CZZ CYS  THIARSAHYDROXY-CYSTEINE
                    'DA2':'ARG',                    ##  DA2 ARG  MODIFIED ARGININE
                    'DAB':'ALA',                    ##  DAB ALA  2,4-DIAMINOBUTYRIC ACID
                    'DAH':'PHE',                    ##  DAH PHE  3,4-DIHYDROXYDAHNYLALANINE
                    'DAL':'ALA',                    ##  DAL ALA  D-ALANINE
                    'DAM':'ALA',                    ##  DAM ALA  N-METHYL-ALPHA-BETA-DEHYDROALANINE
                    'DAR':'ARG',                    ##  DAR ARG  D-ARGININE
                    'DAS':'ASP',                    ##  DAS ASP  D-ASPARTIC ACID
                    'DBU':'ALA',                    ##  DBU ALA  (2E)-2-AMINOBUT-2-ENOIC ACID
                    'DBY':'TYR',                    ##  DBY TYR  3,5 DIBROMOTYROSINE
                    'DBZ':'ALA',                    ##  DBZ ALA  3-(BENZOYLAMINO)-L-ALANINE
                    'DCL':'LEU',                    ##  DCL LEU  2-AMINO-4-METHYL-PENTANYL GROUP
                    'DCY':'CYS',                    ##  DCY CYS  D-CYSTEINE
                    'DDE':'HIS',                    ##  DDE HIS
                    'DGL':'GLU',                    ##  DGL GLU  D-GLU
                    'DGN':'GLN',                    ##  DGN GLN  D-GLUTAMINE
                    'DHA':'ALA',                    ##  DHA ALA  2-AMINO-ACRYLIC ACID
                    'DHI':'HIS',                    ##  DHI HIS  D-HISTIDINE
                    'DHL':'SER',                    ##  DHL SER  POST-TRANSLATIONAL MODIFICATION
                    'DIL':'ILE',                    ##  DIL ILE  D-ISOLEUCINE
                    'DIV':'VAL',                    ##  DIV VAL  D-ISOVALINE
                    'DLE':'LEU',                    ##  DLE LEU  D-LEUCINE
                    'DLS':'LYS',                    ##  DLS LYS  DI-ACETYL-LYSINE
                    'DLY':'LYS',                    ##  DLY LYS  D-LYSINE
                    'DMH':'ASN',                    ##  DMH ASN  N4,N4-DIMETHYL-ASPARAGINE
                    'DMK':'ASP',                    ##  DMK ASP  DIMETHYL ASPARTIC ACID
                    'DNE':'LEU',                    ##  DNE LEU  D-NORLEUCINE
                    'DNG':'LEU',                    ##  DNG LEU  N-FORMYL-D-NORLEUCINE
                    'DNL':'LYS',                    ##  DNL LYS  6-AMINO-HEXANAL
                    'DNM':'LEU',                    ##  DNM LEU  D-N-METHYL NORLEUCINE
                    'DPH':'PHE',                    ##  DPH PHE  DEAMINO-METHYL-PHENYLALANINE
                    'DPL':'PRO',                    ##  DPL PRO  4-OXOPROLINE
                    'DPN':'PHE',                    ##  DPN PHE  D-CONFIGURATION
                    'DPP':'ALA',                    ##  DPP ALA  DIAMMINOPROPANOIC ACID
                    'DPQ':'TYR',                    ##  DPQ TYR  TYROSINE DERIVATIVE
                    'DPR':'PRO',                    ##  DPR PRO  D-PROLINE
                    'DSE':'SER',                    ##  DSE SER  D-SERINE N-METHYLATED
                    'DSG':'ASN',                    ##  DSG ASN  D-ASPARAGINE
                    'DSN':'SER',                    ##  DSN SER  D-SERINE
                    'DTH':'THR',                    ##  DTH THR  D-THREONINE
                    'DTR':'TRP',                    ##  DTR TRP  D-TRYPTOPHAN
                    'DTY':'TYR',                    ##  DTY TYR  D-TYROSINE
                    'DVA':'VAL',                    ##  DVA VAL  D-VALINE
                    'DYG':'ALA',                    ##  DYG ALA
                    'DYS':'CYS',                    ##  DYS CYS
                    'EFC':'CYS',                    ##  EFC CYS  S,S-(2-FLUOROETHYL)THIOCYSTEINE
                    'ESB':'TYR',                    ##  ESB TYR
                    'ESC':'MET',                    ##  ESC MET  2-AMINO-4-ETHYL SULFANYL BUTYRIC ACID
                    'FCL':'PHE',                    ##  FCL PHE  3-CHLORO-L-PHENYLALANINE
                    'FGL':'ALA',                    ##  FGL ALA  2-AMINOPROPANEDIOIC ACID
                    'FGP':'SER',                    ##  FGP SER
                    'FHL':'LYS',                    ##  FHL LYS  MODIFIED LYSINE
                    'FLE':'LEU',                    ##  FLE LEU  FUROYL-LEUCINE
                    'FLT':'TYR',                    ##  FLT TYR  FLUOROMALONYL TYROSINE
                    'FME':'MET',                    ##  FME MET  FORMYL-METHIONINE
                    'FOE':'CYS',                    ##  FOE CYS
                    'FOG':'PHE',                    ##  FOG PHE  PHENYLALANINOYL-[1-HYDROXY]-2-PROPYLENE
                    'FOR':'MET',                    ##  FOR MET
                    'FRF':'PHE',                    ##  FRF PHE  PHE FOLLOWED BY REDUCED PHE
                    'FTR':'TRP',                    ##  FTR TRP  FLUOROTRYPTOPHANE
                    'FTY':'TYR',                    ##  FTY TYR  DEOXY-DIFLUOROMETHELENE-PHOSPHOTYROSINE
                    'GHG':'GLN',                    ##  GHG GLN  GAMMA-HYDROXY-GLUTAMINE
                    'GHP':'GLY',                    ##  GHP GLY  4-HYDROXYPHENYLGLYCINE
                    'GL3':'GLY',                    ##  GL3 GLY  POST-TRANSLATIONAL MODIFICATION
                    'GLH':'GLN',                    ##  GLH GLN
                    'GLN':'GLN',                    ##  GLN GLN
                    'GLU':'GLU',                    ##  GLU GLU
                    'GLY':'GLY',                    ##  GLY GLY
                    'GLZ':'GLY',                    ##  GLZ GLY  AMINO-ACETALDEHYDE
                    'GMA':'GLU',                    ##  GMA GLU  1-AMIDO-GLUTAMIC ACID
                    'GMU':'ALA',                    ##  GMU 5MU
                    'GPL':'LYS',                    ##  GPL LYS  LYSINE GUANOSINE-5'-MONOPHOSPHATE
                    'GT9':'CYS',                    ##  GT9 CYS  SG ALKYLATED
                    'GVL':'SER',                    ##  GVL SER  SERINE MODIFED WITH PHOSPHOPANTETHEINE
                    'GYC':'CYS',                    ##  GYC CYS
                    'GYS':'GLY',                    ##  GYS GLY
                    'H5M':'PRO',                    ##  H5M PRO  TRANS-3-HYDROXY-5-METHYLPROLINE
                    'HHK':'ALA',                    ##  HHK ALA  (2S)-2,8-DIAMINOOCTANOIC ACID
                    'HIA':'HIS',                    ##  HIA HIS  L-HISTIDINE AMIDE
                    'HIC':'HIS',                    ##  HIC HIS  4-METHYL-HISTIDINE
                    'HIP':'HIS',                    ##  HIP HIS  ND1-PHOSPHONOHISTIDINE
                    'HIQ':'HIS',                    ##  HIQ HIS  MODIFIED HISTIDINE
                    'HIS':'HIS',                    ##  HIS HIS
                    'HLU':'LEU',                    ##  HLU LEU  BETA-HYDROXYLEUCINE
                    'HMF':'ALA',                    ##  HMF ALA  2-AMINO-4-PHENYL-BUTYRIC ACID
                    'HMR':'ARG',                    ##  HMR ARG  BETA-HOMOARGININE
                    'HPE':'PHE',                    ##  HPE PHE  HOMOPHENYLALANINE
                    'HPH':'PHE',                    ##  HPH PHE  PHENYLALANINOL GROUP
                    'HPQ':'PHE',                    ##  HPQ PHE  HOMOPHENYLALANINYLMETHANE
                    'HRG':'ARG',                    ##  HRG ARG  L-HOMOARGININE
                    'HSE':'SER',                    ##  HSE SER  L-HOMOSERINE
                    'HSL':'SER',                    ##  HSL SER  HOMOSERINE LACTONE
                    'HSO':'HIS',                    ##  HSO HIS  HISTIDINOL
                    'HTI':'CYS',                    ##  HTI CYS
                    'HTR':'TRP',                    ##  HTR TRP  BETA-HYDROXYTRYPTOPHANE
                    'HY3':'PRO',                    ##  HY3 PRO  3-HYDROXYPROLINE
                    'HYP':'PRO',                    ##  HYP PRO  4-HYDROXYPROLINE
                    'IAM':'ALA',                    ##  IAM ALA  4-[(ISOPROPYLAMINO)METHYL]PHENYLALANINE
                    'IAS':'ASP',                    ##  IAS ASP  ASPARTYL GROUP
                    'IGL':'ALA',                    ##  IGL ALA  ALPHA-AMINO-2-INDANACETIC ACID
                    'IIL':'ILE',                    ##  IIL ILE  ISO-ISOLEUCINE
                    'ILE':'ILE',                    ##  ILE ILE
                    'ILG':'GLU',                    ##  ILG GLU  GLU LINKED TO NEXT RESIDUE VIA CG
                    'ILX':'ILE',                    ##  ILX ILE  4,5-DIHYDROXYISOLEUCINE
                    'IML':'ILE',                    ##  IML ILE  N-METHYLATED
                    'IPG':'GLY',                    ##  IPG GLY  N-ISOPROPYL GLYCINE
                    'IT1':'LYS',                    ##  IT1 LYS
                    'IYR':'TYR',                    ##  IYR TYR  3-IODO-TYROSINE
                    'KCX':'LYS',                    ##  KCX LYS  CARBAMOYLATED LYSINE
                    'KGC':'LYS',                    ##  KGC LYS
                    'KOR':'CYS',                    ##  KOR CYS  MODIFIED CYSTEINE
                    'KST':'LYS',                    ##  KST LYS  N~6~-(5-CARBOXY-3-THIENYL)-L-LYSINE
                    'KYN':'ALA',                    ##  KYN ALA  KYNURENINE
                    'LA2':'LYS',                    ##  LA2 LYS
                    'LAL':'ALA',                    ##  LAL ALA  N,N-DIMETHYL-L-ALANINE
                    'LCK':'LYS',                    ##  LCK LYS
                    'LCX':'LYS',                    ##  LCX LYS  CARBAMYLATED LYSINE
                    'LDH':'LYS',                    ##  LDH LYS  N~6~-ETHYL-L-LYSINE
                    'LED':'LEU',                    ##  LED LEU  POST-TRANSLATIONAL MODIFICATION
                    'LEF':'LEU',                    ##  LEF LEU  2-5-FLUOROLEUCINE
                    'LET':'LYS',                    ##  LET LYS  ODIFIED LYSINE
                    'LEU':'LEU',                    ##  LEU LEU
                    'LLP':'LYS',                    ##  LLP LYS
                    'LLY':'LYS',                    ##  LLY LYS  NZ-(DICARBOXYMETHYL)LYSINE
                    'LME':'GLU',                    ##  LME GLU  (3R)-3-METHYL-L-GLUTAMIC ACID
                    'LNT':'LEU',                    ##  LNT LEU
                    'LPD':'PRO',                    ##  LPD PRO  L-PROLINAMIDE
                    'LSO':'LYS',                    ##  LSO LYS  MODIFIED LYSINE
                    'LYM':'LYS',                    ##  LYM LYS  DEOXY-METHYL-LYSINE
                    'LYN':'LYS',                    ##  LYN LYS  2,6-DIAMINO-HEXANOIC ACID AMIDE
                    'LYP':'LYS',                    ##  LYP LYS  N~6~-METHYL-N~6~-PROPYL-L-LYSINE
                    'LYR':'LYS',                    ##  LYR LYS  MODIFIED LYSINE
                    'LYS':'LYS',                    ##  LYS LYS
                    'LYX':'LYS',                    ##  LYX LYS  N''-(2-COENZYME A)-PROPANOYL-LYSINE
                    'LYZ':'LYS',                    ##  LYZ LYS  5-HYDROXYLYSINE
                    'M0H':'CYS',                    ##  M0H CYS  S-(HYDROXYMETHYL)-L-CYSTEINE
                    'M2L':'LYS',                    ##  M2L LYS
                    'M3L':'LYS',                    ##  M3L LYS  N-TRIMETHYLLYSINE
                    'MAA':'ALA',                    ##  MAA ALA  N-METHYLALANINE
                    'MAI':'ARG',                    ##  MAI ARG  DEOXO-METHYLARGININE
                    'MBQ':'TYR',                    ##  MBQ TYR
                    'MC1':'SER',                    ##  MC1 SER  METHICILLIN ACYL-SERINE
                    'MCL':'LYS',                    ##  MCL LYS  NZ-(1-CARBOXYETHYL)-LYSINE
                    'MCS':'CYS',                    ##  MCS CYS  MALONYLCYSTEINE
                    'MDO':'ALA',                    ##  MDO ALA
                    'MEA':'PHE',                    ##  MEA PHE  N-METHYLPHENYLALANINE
                    'MEG':'GLU',                    ##  MEG GLU  (2S,3R)-3-METHYL-GLUTAMIC ACID
                    'MEN':'ASN',                    ##  MEN ASN  GAMMA METHYL ASPARAGINE
                    'MET':'MET',                    ##  MET MET
                    'MEU':'GLY',                    ##  MEU GLY  O-METHYL-GLYCINE
                    'MFC':'ALA',                    ##  MFC ALA  CYCLIZED
                    'MGG':'ARG',                    ##  MGG ARG  MODIFIED D-ARGININE
                    'MGN':'GLN',                    ##  MGN GLN  2-METHYL-GLUTAMINE
                    'MHL':'LEU',                    ##  MHL LEU  N-METHYLATED, HYDROXY
                    'MHO':'MET',                    ##  MHO MET  POST-TRANSLATIONAL MODIFICATION
                    'MHS':'HIS',                    ##  MHS HIS  1-N-METHYLHISTIDINE
                    'MIS':'SER',                    ##  MIS SER  MODIFIED SERINE
                    'MLE':'LEU',                    ##  MLE LEU  N-METHYLATED
                    'MLL':'LEU',                    ##  MLL LEU  METHYL L-LEUCINATE
                    'MLY':'LYS',                    ##  MLY LYS  METHYLATED LYSINE
                    'MLZ':'LYS',                    ##  MLZ LYS  N-METHYL-LYSINE
                    'MME':'MET',                    ##  MME MET  N-METHYL METHIONINE
                    'MNL':'LEU',                    ##  MNL LEU  4,N-DIMETHYLNORLEUCINE
                    'MNV':'VAL',                    ##  MNV VAL  N-METHYL-C-AMINO VALINE
                    'MPQ':'GLY',                    ##  MPQ GLY  N-METHYL-ALPHA-PHENYL-GLYCINE
                    'MSA':'GLY',                    ##  MSA GLY  (2-S-METHYL) SARCOSINE
                    'MSE':'MET',                    ##  MSE MET  ELENOMETHIONINE
                    'MSO':'MET',                    ##  MSO MET  METHIONINE SULFOXIDE
                    'MTY':'PHE',                    ##  MTY PHE  3-HYDROXYPHENYLALANINE
                    'MVA':'VAL',                    ##  MVA VAL  N-METHYLATED
                    'N10':'SER',                    ##  N10 SER  O-[(HEXYLAMINO)CARBONYL]-L-SERINE
                    'NAL':'ALA',                    ##  NAL ALA  BETA-(2-NAPHTHYL)-ALANINE
                    'NAM':'ALA',                    ##  NAM ALA  NAM NAPTHYLAMINOALANINE
                    'NBQ':'TYR',                    ##  NBQ TYR
                    'NC1':'SER',                    ##  NC1 SER  NITROCEFIN ACYL-SERINE
                    'NCB':'ALA',                    ##  NCB ALA  CHEMICAL MODIFICATION
                    'NEP':'HIS',                    ##  NEP HIS  N1-PHOSPHONOHISTIDINE
                    'NFA':'PHE',                    ##  NFA PHE  MODIFIED PHENYLALANINE
                    'NIY':'TYR',                    ##  NIY TYR  META-NITRO-TYROSINE
                    'NLE':'LEU',                    ##  NLE LEU  NORLEUCINE
                    'NLN':'LEU',                    ##  NLN LEU  NORLEUCINE AMIDE
                    'NLO':'LEU',                    ##  NLO LEU  O-METHYL-L-NORLEUCINE
                    'NMC':'GLY',                    ##  NMC GLY  N-CYCLOPROPYLMETHYL GLYCINE
                    'NMM':'ARG',                    ##  NMM ARG  MODIFIED ARGININE
                    'NPH':'CYS',                    ##  NPH CYS
                    'NRQ':'ALA',                    ##  NRQ ALA
                    'NVA':'VAL',                    ##  NVA VAL  NORVALINE
                    'NYC':'ALA',                    ##  NYC ALA
                    'NYS':'CYS',                    ##  NYS CYS
                    'NZH':'HIS',                    ##  NZH HIS
                    'OAS':'SER',                    ##  OAS SER  O-ACETYLSERINE
                    'OBS':'LYS',                    ##  OBS LYS  MODIFIED LYSINE
                    'OCS':'CYS',                    ##  OCS CYS  CYSTEINE SULFONIC ACID
                    'OCY':'CYS',                    ##  OCY CYS  HYDROXYETHYLCYSTEINE
                    'OHI':'HIS',                    ##  OHI HIS  3-(2-OXO-2H-IMIDAZOL-4-YL)-L-ALANINE
                    'OHS':'ASP',                    ##  OHS ASP  O-(CARBOXYSULFANYL)-4-OXO-L-HOMOSERINE
                    'OLT':'THR',                    ##  OLT THR  O-METHYL-L-THREONINE
                    'OMT':'MET',                    ##  OMT MET  METHIONINE SULFONE
                    'OPR':'ARG',                    ##  OPR ARG  C-(3-OXOPROPYL)ARGININE
                    'ORN':'ALA',                    ##  ORN ALA  ORNITHINE
                    'ORQ':'ARG',                    ##  ORQ ARG  N~5~-ACETYL-L-ORNITHINE
                    'OSE':'SER',                    ##  OSE SER  O-SULFO-L-SERINE
                    'OTY':'TYR',                    ##  OTY TYR
                    'OXX':'ASP',                    ##  OXX ASP  OXALYL-ASPARTYL ANHYDRIDE
                    'P1L':'CYS',                    ##  P1L CYS  S-PALMITOYL CYSTEINE
                    'P2Y':'PRO',                    ##  P2Y PRO  (2S)-PYRROLIDIN-2-YLMETHYLAMINE
                    'PAQ':'TYR',                    ##  PAQ TYR  SEE REMARK 999
                    'PAT':'TRP',                    ##  PAT TRP  ALPHA-PHOSPHONO-TRYPTOPHAN
                    'PBB':'CYS',                    ##  PBB CYS  S-(4-BROMOBENZYL)CYSTEINE
                    'PBF':'PHE',                    ##  PBF PHE  PARA-(BENZOYL)-PHENYLALANINE
                    'PCA':'PRO',                    ##  PCA PRO  5-OXOPROLINE
                    'PCS':'PHE',                    ##  PCS PHE  PHENYLALANYLMETHYLCHLORIDE
                    'PEC':'CYS',                    ##  PEC CYS  S,S-PENTYLTHIOCYSTEINE
                    'PF5':'PHE',                    ##  PF5 PHE  2,3,4,5,6-PENTAFLUORO-L-PHENYLALANINE
                    'PFF':'PHE',                    ##  PFF PHE  4-FLUORO-L-PHENYLALANINE
                    'PG1':'SER',                    ##  PG1 SER  BENZYLPENICILLOYL-ACYLATED SERINE
                    'PG9':'GLY',                    ##  PG9 GLY  D-PHENYLGLYCINE
                    'PHA':'PHE',                    ##  PHA PHE  PHENYLALANINAL
                    'PHD':'ASP',                    ##  PHD ASP  2-AMINO-4-OXO-4-PHOSPHONOOXY-BUTYRIC ACID
                    'PHE':'PHE',                    ##  PHE PHE
                    'PHI':'PHE',                    ##  PHI PHE  IODO-PHENYLALANINE
                    'PHL':'PHE',                    ##  PHL PHE  L-PHENYLALANINOL
                    'PHM':'PHE',                    ##  PHM PHE  PHENYLALANYLMETHANE
                    'PIA':'ALA',                    ##  PIA ALA  FUSION OF ALA 65, TYR 66, GLY 67
                    'PLE':'LEU',                    ##  PLE LEU  LEUCINE PHOSPHINIC ACID
                    'PM3':'PHE',                    ##  PM3 PHE
                    'POM':'PRO',                    ##  POM PRO  CIS-5-METHYL-4-OXOPROLINE
                    'PPH':'LEU',                    ##  PPH LEU  PHENYLALANINE PHOSPHINIC ACID
                    'PPN':'PHE',                    ##  PPN PHE  THE LIGAND IS A PARA-NITRO-PHENYLALANINE
                    'PR3':'CYS',                    ##  PR3 CYS  INE DTT-CYSTEINE
                    'PRO':'PRO',                    ##  PRO PRO
                    'PRQ':'PHE',                    ##  PRQ PHE  PHENYLALANINE
                    'PRR':'ALA',                    ##  PRR ALA  3-(METHYL-PYRIDINIUM)ALANINE
                    'PRS':'PRO',                    ##  PRS PRO  THIOPROLINE
                    'PSA':'PHE',                    ##  PSA PHE
                    'PSH':'HIS',                    ##  PSH HIS  1-THIOPHOSPHONO-L-HISTIDINE
                    'PTH':'TYR',                    ##  PTH TYR  METHYLENE-HYDROXY-PHOSPHOTYROSINE
                    'PTM':'TYR',                    ##  PTM TYR  ALPHA-METHYL-O-PHOSPHOTYROSINE
                    'PTR':'TYR',                    ##  PTR TYR  O-PHOSPHOTYROSINE
                    'PYA':'ALA',                    ##  PYA ALA  3-(1,10-PHENANTHROL-2-YL)-L-ALANINE
                    'PYC':'ALA',                    ##  PYC ALA  PYRROLE-2-CARBOXYLATE
                    'PYR':'SER',                    ##  PYR SER  CHEMICALLY MODIFIED
                    'PYT':'ALA',                    ##  PYT ALA  MODIFIED ALANINE
                    'PYX':'CYS',                    ##  PYX CYS  S-[S-THIOPYRIDOXAMINYL]CYSTEINE
                    'R1A':'CYS',                    ##  R1A CYS
                    'R1B':'CYS',                    ##  R1B CYS
                    'R1F':'CYS',                    ##  R1F CYS
                    'R7A':'CYS',                    ##  R7A CYS
                    'RC7':'ALA',                    ##  RC7 ALA
                    'RCY':'CYS',                    ##  RCY CYS
                    'S1H':'SER',                    ##  S1H SER  1-HEXADECANOSULFONYL-O-L-SERINE
                    'SAC':'SER',                    ##  SAC SER  N-ACETYL-SERINE
                    'SAH':'CYS',                    ##  SAH CYS  S-ADENOSYL-L-HOMOCYSTEINE
                    'SAR':'GLY',                    ##  SAR GLY  SARCOSINE
                    'SBD':'SER',                    ##  SBD SER
                    'SBG':'SER',                    ##  SBG SER  MODIFIED SERINE
                    'SBL':'SER',                    ##  SBL SER
                    'SC2':'CYS',                    ##  SC2 CYS  N-ACETYL-L-CYSTEINE
                    'SCH':'CYS',                    ##  SCH CYS  S-METHYL THIOCYSTEINE GROUP
                    'SCS':'CYS',                    ##  SCS CYS  MODIFIED CYSTEINE
                    'SCY':'CYS',                    ##  SCY CYS  CETYLATED CYSTEINE
                    'SDP':'SER',                    ##  SDP SER
                    'SEB':'SER',                    ##  SEB SER  O-BENZYLSULFONYL-SERINE
                    'SEC':'ALA',                    ##  SEC ALA  2-AMINO-3-SELENINO-PROPIONIC ACID
                    'SEL':'SER',                    ##  SEL SER  2-AMINO-1,3-PROPANEDIOL
                    'SEP':'SER',                    ##  SEP SER  E PHOSPHOSERINE
                    'SER':'SER',                    ##  SER SER
                    'SET':'SER',                    ##  SET SER  AMINOSERINE
                    'SGB':'SER',                    ##  SGB SER  MODIFIED SERINE
                    'SGR':'SER',                    ##  SGR SER  MODIFIED SERINE
                    'SHC':'CYS',                    ##  SHC CYS  S-HEXYLCYSTEINE
                    'SHP':'GLY',                    ##  SHP GLY  (4-HYDROXYMALTOSEPHENYL)GLYCINE
                    'SIC':'ALA',                    ##  SIC ALA
                    'SLZ':'LYS',                    ##  SLZ LYS  L-THIALYSINE
                    'SMC':'CYS',                    ##  SMC CYS  POST-TRANSLATIONAL MODIFICATION
                    'SME':'MET',                    ##  SME MET  METHIONINE SULFOXIDE
                    'SMF':'PHE',                    ##  SMF PHE  4-SULFOMETHYL-L-PHENYLALANINE
                    'SNC':'CYS',                    ##  SNC CYS  S-NITROSO CYSTEINE
                    'SNN':'ASP',                    ##  SNN ASP  POST-TRANSLATIONAL MODIFICATION
                    'SOC':'CYS',                    ##  SOC CYS  DIOXYSELENOCYSTEINE
                    'SOY':'SER',                    ##  SOY SER  OXACILLOYL-ACYLATED SERINE
                    'SUI':'ALA',                    ##  SUI ALA
                    'SUN':'SER',                    ##  SUN SER  TABUN CONJUGATED SERINE
                    'SVA':'SER',                    ##  SVA SER  SERINE VANADATE
                    'SVV':'SER',                    ##  SVV SER  MODIFIED SERINE
                    'SVX':'SER',                    ##  SVX SER  MODIFIED SERINE
                    'SVY':'SER',                    ##  SVY SER  MODIFIED SERINE
                    'SVZ':'SER',                    ##  SVZ SER  MODIFIED SERINE
                    'SXE':'SER',                    ##  SXE SER  MODIFIED SERINE
                    'TBG':'GLY',                    ##  TBG GLY  T-BUTYL GLYCINE
                    'TBM':'THR',                    ##  TBM THR
                    'TCQ':'TYR',                    ##  TCQ TYR  MODIFIED TYROSINE
                    'TEE':'CYS',                    ##  TEE CYS  POST-TRANSLATIONAL MODIFICATION
                    'TH5':'THR',                    ##  TH5 THR  O-ACETYL-L-THREONINE
                    'THC':'THR',                    ##  THC THR  N-METHYLCARBONYLTHREONINE
                    'THR':'THR',                    ##  THR THR
                    'TIH':'ALA',                    ##  TIH ALA  BETA(2-THIENYL)ALANINE
                    'TMD':'THR',                    ##  TMD THR  N-METHYLATED, EPSILON C ALKYLATED
                    'TNB':'CYS',                    ##  TNB CYS  S-(2,3,6-TRINITROPHENYL)CYSTEINE
                    'TOX':'TRP',                    ##  TOX TRP
                    'TPL':'TRP',                    ##  TPL TRP  TRYTOPHANOL
                    'TPO':'THR',                    ##  TPO THR  HOSPHOTHREONINE
                    'TPQ':'ALA',                    ##  TPQ ALA  2,4,5-TRIHYDROXYPHENYLALANINE
                    'TQQ':'TRP',                    ##  TQQ TRP
                    'TRF':'TRP',                    ##  TRF TRP  N1-FORMYL-TRYPTOPHAN
                    'TRN':'TRP',                    ##  TRN TRP  AZA-TRYPTOPHAN
                    'TRO':'TRP',                    ##  TRO TRP  2-HYDROXY-TRYPTOPHAN
                    'TRP':'TRP',                    ##  TRP TRP
                    'TRQ':'TRP',                    ##  TRQ TRP
                    'TRW':'TRP',                    ##  TRW TRP
                    'TRX':'TRP',                    ##  TRX TRP  6-HYDROXYTRYPTOPHAN
                    'TTQ':'TRP',                    ##  TTQ TRP  6-AMINO-7-HYDROXY-L-TRYPTOPHAN
                    'TTS':'TYR',                    ##  TTS TYR
                    'TY2':'TYR',                    ##  TY2 TYR  3-AMINO-L-TYROSINE
                    'TY3':'TYR',                    ##  TY3 TYR  3-HYDROXY-L-TYROSINE
                    'TYB':'TYR',                    ##  TYB TYR  TYROSINAL
                    'TYC':'TYR',                    ##  TYC TYR  L-TYROSINAMIDE
                    'TYI':'TYR',                    ##  TYI TYR  3,5-DIIODOTYROSINE
                    'TYN':'TYR',                    ##  TYN TYR  ADDUCT AT HYDROXY GROUP
                    'TYO':'TYR',                    ##  TYO TYR
                    'TYQ':'TYR',                    ##  TYQ TYR  AMINOQUINOL FORM OF TOPA QUINONONE
                    'TYR':'TYR',                    ##  TYR TYR
                    'TYS':'TYR',                    ##  TYS TYR  INE SULPHONATED TYROSINE
                    'TYT':'TYR',                    ##  TYT TYR
                    'TYX':'CYS',                    ##  TYX CYS  S-(2-ANILINO-2-OXOETHYL)-L-CYSTEINE
                    'TYY':'TYR',                    ##  TYY TYR  IMINOQUINONE FORM OF TOPA QUINONONE
                    'TYZ':'ARG',                    ##  TYZ ARG  PARA ACETAMIDO BENZOIC ACID
                    'UMA':'ALA',                    ##  UMA ALA
                    'VAD':'VAL',                    ##  VAD VAL  DEAMINOHYDROXYVALINE
                    'VAF':'VAL',                    ##  VAF VAL  METHYLVALINE
                    'VAL':'VAL',                    ##  VAL VAL
                    'VDL':'VAL',                    ##  VDL VAL  (2R,3R)-2,3-DIAMINOBUTANOIC ACID
                    'VLL':'VAL',                    ##  VLL VAL  (2S)-2,3-DIAMINOBUTANOIC ACID
                    'HSD':'HIS',                    ##  up his
                    'VME':'VAL',                    ##  VME VAL  O- METHYLVALINE
                    'X9Q':'ALA',                    ##  X9Q ALA
                    'XX1':'LYS',                    ##  XX1 LYS  N~6~-7H-PURIN-6-YL-L-LYSINE
                    'XXY':'ALA',                    ##  XXY ALA
                    'XYG':'ALA',                    ##  XYG ALA
                    'YCM':'CYS',                    ##  YCM CYS  S-(2-AMINO-2-OXOETHYL)-L-CYSTEINE
                    'YOF':'TYR' }                   ##  YOF TYR  3-FLUOROTYROSINE
    nonstandard = modres.keys()





# -*- coding: utf-8 -*-

import os
from os.path import join, abspath
from subprocess import PIPE, Popen
from pkg_resources import resource_filename
from aggrescan_3d import run as run_aggrescan_3d
import logger
import platform

_name = "Analysis"
if platform.system() == "Windows":
    _bin_name = "freesasa.exe"
elif platform.system() == "Darwin":
    _bin_name = "freesasa_Darwin"
else:
    _bin_name = "freesasa"


def aggregation_analysis(config, target="input.pdb", working_dir=".", agg_work_dir=""):
    """Takes Aggrescaan3D Job class config as an argument and should never change it"""
    os.chdir(working_dir)  # Just in case one of the programs handles abspaths poorly
    if config["naccess"]:
        _run_naccess(target=target, agg=agg_work_dir, naccess_dir=config["naccess"])
    else:
        _run_free_sasa(target=target, agg=agg_work_dir)
    _run_aggrescan(target=target, working_dir=working_dir, config=config)


def _run_aggrescan(target, working_dir, config):
    """
    Takes target pdb file, working directory for analysis
    and aggrescan config dict as arguments so need the two in sync
    """
    logger.info(module_name=_name, msg="Starting Aggrescan3D on %s" % target.split("/")[-1])
    logger.debug(module_name=_name, msg="Running Aggrescan3D program on %s" % abspath(target))
    matrix_dir = resource_filename('aggrescan', join("data", "matrices", "aggrescan.mat"))
    out = run_aggrescan_3d(target, matrix_dir, config["distance"], working_dir, naccess=config["naccess"])
    logger.to_file(filename=join(config["tmp_dir"], "input_output"), content=out, allow_err=False)


def _run_free_sasa(target="folded.pdb", agg=""):
    sasa_dir = resource_filename('aggrescan', join("data", "freesasa-2.0.1", "src", _bin_name))
    sasa_cmd = [sasa_dir, "--resolution", "100", target, "--radii", "naccess", "--format", "pdb", "--format", "rsa"]
    logger.debug(module_name="freeSasa", msg="Starting freeSasa with %s" % " ".join(sasa_cmd))
    _fname = join(agg, "ASA.error")
    out, err = Popen(sasa_cmd, stdout=PIPE, stderr=PIPE).communicate()
    if err:
        logger.to_file(_fname, content=err)
        for line in err.split("\n"):
            if "warning" in line:
                logger.critical(module_name="freeSasa", msg="Warning detected: %s" % line)
                logger.critical(module_name="freeSasa", msg="Attempting to continue.")
            else:
                raise logger.ASAError(program_name="freeSasa", filename=_fname)
    logger.to_file(filename="sasa.out", content=out, allow_err=False)


def _run_naccess(target="folded.pdb", agg="", naccess_dir=""):
    naccess_cmd = [naccess_dir, target, "-p", "1.4", "-z", "0.05"]
    logger.debug(module_name=_name, msg="running naccess program with %s" % " ".join(naccess_cmd))
    _fname = join(agg, "ASA.error")
    try:
        out, err = Popen(naccess_cmd, stdout=PIPE, stderr=PIPE).communicate()
    except OSError:
        raise logger.AggrescanError("Couldn't start the naccess program. Naccess requires csh shell to be present on the system."
                                    " If it is present please run the program in debug mode and see if the path printed in "
                                    "Naccess command actually contains the file.", module_name="Naccess")
    if err:
        logger.to_file(filename=_fname, content=err)
        raise logger.ASAError(program_name="Naccess", filename=_fname)
    try:
        for msg in out.split("\n"):
            if msg:
                logger.debug(module_name="Naccess", msg=msg)
    except AttributeError:
        pass




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import multiprocessing as mp
from subprocess import Popen, PIPE
from collections import OrderedDict
from os.path import join, isfile
from postProcessing import make_auto_mut_plot
import os
import re
import shutil
import logger

_name = "Auto_mut"

# Define some variables that dictate how some properties of this - maybe they get to be parameters in the future
_score_threshold = -0.2  # How high an A3D score has to be for the residue to be considered for mutation
_energy_threshold = 0.0   # When a mutation is considered "good"
_score_diff_threshold = 5  # When a mutation is considered to increase solubility  +5 should accept all while 0 would
                           # already be qutite restrictive
_target_mutations = ["E", "K", "D", "R"]  # Glutamic acid, lysine,  aspartic acid, arginine

# The slicing in this code can be hard to read so here is a short summary:
# Chain ID and residue names come as one letter shorts while the number can have more than one char
# Final mutation code is <Old residue><New residue><Residue number><Chain ID>
# The auto_mutation to_exclude argument comes as <Residue number><Chain ID>
# The one that comes from A3D file comes as < Old residue><Residue number><Chain ID>


def run_auto_mutation(work_dir, options, foldx_loc, distance):
    n_mutations = options[0]
    n_processes = options[1]
    to_exclude = []     # Not to worry about it even exists
    if len(options) > 2:
        to_exclude = options[2]
    pool = mp.Pool(n_processes)
    mutations, avg_score = _mutation_list(work_dir=work_dir, excluded_list=to_exclude, n_mutations=n_mutations)
    if not mutations:
        with open(join(work_dir, "Mutations_summary.csv"), "w") as f:   # leave an empty file for the server
            pass
        return
    pool.map(_run_job, [(i, foldx_loc, work_dir, str(distance)) for i in mutations])

    _analyze_results(work_dir=work_dir, output_file="Mutations_summary.csv", mutation_list=mutations,
                     base_avg_score=avg_score)
    _cleanup(work_dir=work_dir, mutation_list=mutations)
    try:
        _plots(work_dir=work_dir)
    except Exception as e:  # This is hopefully not needed but in case something happens the user will at least see a
                            # message rather than a traceback
        logger.critical(module_name=_name, msg="It seems that all the mutation attempts failed or some other unexpected"
                                               " error arisen while trying to plot the automated mutations.")
        raise


def _mutation_list(work_dir, excluded_list, n_mutations):
    scores = _parse_a3dcsv(os.path.join(work_dir, "A3D.csv"))
    avg_score = sum(scores.values())/len(scores.values())
    mutation_list = []
    counter = 0
    for residue, value in scores.items():
        if value > _score_threshold and residue[1:] not in excluded_list and residue[0] not in _target_mutations \
                and value != 0:
            mutation_list.extend(["%s%s%s" % (residue[0], i, residue[1:]) for i in _target_mutations])
            logger.info(module_name=_name, msg="Residue number %s from chain %s and a score of %.3f (%s) selected "
                                               "for automated muatation" % (residue[1:-1], residue[-1], value,
                                                                            _aa_dict_F[residue[0]]))
            counter += 1
            if counter >= n_mutations:
                break
        elif value > _score_threshold and residue[1:] in excluded_list:
            logger.info(module_name=_name, msg="Residue number %s from chain %s and a score of %.3f omitted "
                                               "from automated muatation (excluded by the user)." % (residue[1:-1], residue[-1], value))
    if not mutation_list:
        logger.critical(module_name=_name, msg="Couldn't find residues suitable for automated mutations (exceeding a "
                                               "threshold of %.2f). No automated mutations performed." % _score_threshold)
    return mutation_list, avg_score


def _parse_a3dcsv(filepath):  #TODO this is done on muttiple occasions so maybe should be unified somwhere
    """
    Return an OrderedDict of label:score type. The dict is sorted by score so highest is on top
    """
    pattern = re.compile(r"^(.*),(.*),(.*),(.*),(.*)$", re.M)
    scores = OrderedDict()
    try:
        with open(filepath, 'r') as f:
            data = pattern.findall(f.read().replace("\r", ""))[1:]  #
    except IOError:
        return False    # The mutation likely failed this should pass the info to analyze_results
    for line in data:
        label = line[3] + line[2] + line[1]  # One letter code + residue ID + chain ID (the mutation syntax)
        aggScore = float(line[4])
        scores[label] = aggScore
    scores = OrderedDict(sorted(scores.items(), key=lambda x: x[1], reverse=True))
    return scores


def _run_job(args):
    """
    Run a single A3D job with a specific mutation
    args go as follows: mutation code, FoldX location, main job's work dir, a3d distance argument
    """
    mutation, foldx, work_dir, distance = args
    os.chdir(work_dir)
    cmd = ["aggrescan", "-i", "output.pdb", "-v", "4", "-w", mutation, "-m", mutation, "-f", foldx,
           "--subprocess", "--distance", distance]
    logger.info(module_name=_name, msg="Mutating residue number %s from chain %s (%s) into %s "
                                       " " % (mutation[2:-1], mutation[-1], _aa_dict_F[mutation[0]],
                                              _aa_dict_F[mutation[1]]))  # converting letters into full names
    proc = Popen(cmd, stdout=PIPE, stderr=PIPE)
    proc.communicate()
    if proc.returncode != 0:
        logger.warning(module_name=_name, msg="Mutation %s could have failed (this can be ignored if the main program "
                                              "reports the energy difference). Simulation log for that run should be "
                                               "available at %s" % (mutation,
                                                                    os.path.join(work_dir, mutation, "Aggrescan.error")))


def _analyze_results(work_dir, output_file, mutation_list, base_avg_score):
    """
    Analyze the results and select all of those that are not relevant on keeping top X mutations and return the rest
    to the cleaner that will get rid of them, but keeping their scores in the output_file
    """
    data = OrderedDict()
    unnecessary_results = []
    for mutation in mutation_list:
        scores = _parse_a3dcsv(os.path.join(work_dir, mutation, "A3D.csv"))
        if not scores:
            continue
        with open(os.path.join(work_dir, mutation, "MutantEnergyDiff"), 'r') as f:
            mutation_energy = float(f.read().split()[0])    # This should be guaranteed to work given the check above
        avg_score = sum(scores.values())/len(scores.values())
        data[mutation] = [mutation_energy, avg_score, avg_score - base_avg_score]
        if mutation_energy > _energy_threshold or avg_score - base_avg_score > _score_diff_threshold:
            unnecessary_results.append(mutation)
        logger.info(module_name=_name, msg="Effect of mutation residue number %s from chain %s (%s) into %s: "
                                           "Energy difference: %.4f kcal/mol, Difference in average score from the "
                                           "base case: %.4f"
                                           "" % (mutation[2:-1], mutation[-1], _aa_dict_F[mutation[0]],
                                                 _aa_dict_F[mutation[1]], mutation_energy, avg_score - base_avg_score))
    data = OrderedDict(sorted(data.items(), key=lambda x: x[1][0]))   # sort by mutation energy
    with open(os.path.join(work_dir, output_file), "w") as f:
        f.write("%s,%s,%s,%s\n" % ("Mutation", "EnergyDiff", "AvgScore", "AvgScoreDiff"))
        for mutation, values in data.items():
            f.write("%s,%.4f,%.4f,%.4f\n" % (mutation, values[0], values[1], values[2]))
    return unnecessary_results


def _cleanup(work_dir, mutation_list):
    for mutation in mutation_list:
        if isfile(join(work_dir, mutation, "A3D.csv")) and isfile(join(work_dir, mutation, "output.pdb")):
            shutil.move(join(work_dir, mutation, "A3D.csv"), join(work_dir, "%s%s" % (mutation, ".csv")))
            shutil.move(join(work_dir, mutation, "output.pdb"), join(work_dir, "%s%s" % (mutation, ".pdb")))
            shutil.rmtree(join(work_dir, mutation))
        else:
            if isfile(join(work_dir, mutation, "Aggrescan.error")):
                shutil.move(join(work_dir, mutation, "Aggrescan.error"), join(work_dir, "%s%s" % (mutation, ".error")))
                shutil.rmtree(join(work_dir, mutation))
            else:
                with open(join(work_dir, "%s%s" %(mutation, ".error")), "w") as f:
                    f.write("The mutation has failed and no error log was created during the simulation. "
                            "This is unexpected and if you require further assistance please contact us or leave a bug "
                            "report on our bitbucket at "
                            "https://bitbucket.org/lcbio/aggrescan3d/issues?status=new&status=open")


def _plots(work_dir):
    make_auto_mut_plot(work_dir)


# This is a copy from somewhere else, maybe should put it somewhere for imports
_aa_dict_F = {'A': 'alanine', 'R': 'arginine', 'N': 'asparagine',
              'D': 'aspartic acid', 'C': 'cysteine', 'E': 'glutamic acid',
              'Q': 'glutamine', 'G': 'glycine', 'H': 'histidine',
              'I': 'isoleucine', 'L': 'leucine', 'K': 'lysine',
              'M': 'methionine', 'F': 'phenylalanine', 'P': 'proline',
              'S': 'serine', 'T': 'threonine', 'W': 'tryptophan',
              'Y': 'tyrosine', 'V': 'valine', 'X': 'unknown'}


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from a3d_gui import app
from flask import render_template, g, request, Response, send_from_directory, abort, jsonify, url_for
import os
from utils import query_db
from db_manager import get_filepath
from server_plot import create_plot, create_mut_plot


@app.teardown_appcontext
def close_db(error):
    """Closes the database again at the end of the request."""
    if hasattr(g, 'sqlite_db'):
        g.sqlite_db.close()


@app.errorhandler(404)
def page_not_found(error):
    app.logger.warning('Page not found (IP: '+request.remote_addr+') '+request.url)
    return render_template('page_not_found.html', code="404"), 404


@app.errorhandler(500)
def page_error(error):
    app.logger.warning('Page not found (IP: '+request.remote_addr+') '+request.url)
    return render_template('page_not_found.html', code="500"), 500


# This function fixes a mismatch between new csv format and interactive js plotting
# Its here becouse I dont want to mess with js
@app.route('/compute_static/<jobid>/a3d.csv')
def compute_static_a3dtable(jobid):
    o = ""
    with open(get_filepath(jobid, "A3D.csv")) as fw:
        for line in fw:
            line = line.replace("chain", "ch")
            line = line.replace("residue_name", "resn")
            line = line.replace("residue", "resi")
            o += line.replace("protein,", "").replace("folded,", "")
        return Response(o, status=200, mimetype='text/csv')


@app.route('/compute_static/<jobid>/<folder>/<filename>')
def ustatic_folder(jobid, folder, filename):
    data_dir = query_db("SELECT working_dir FROM user_queue WHERE \
            jid=?", [jobid], one=True)[0]
    d = os.path.abspath(os.path.join(data_dir, folder))
    s = os.path.join(d, filename)
    if os.path.exists(s):
        return send_from_directory(d, filename)
    else:
        abort(404)

@app.route('/compute_static/<jobid>/<filename>')
def ustatic(jobid, filename):
    data_dir = query_db("SELECT working_dir FROM user_queue WHERE \
            jid=?", [jobid], one=True)[0]
    s = os.path.join(data_dir, filename)
    if os.path.exists(s):
        return send_from_directory(data_dir, filename)
    else:
        abort(404)


@app.route('/save_picture/<jobid>', methods=['POST'])
def save_uri(jobid):
    try:
        picture = request.json['file_content'].split(',')[1]
        filename = request.json['filename']
        pic_type = request.json['type']
        model = request.json['model']
        while os.path.isfile(os.path.join(app.config['PICTURES'], filename)):
            # TODO this will be bad if more dots, fix maybe
            filename = filename.split(".")[0] + "_1" + "." + filename.split(".")[-1]   # Prevent overwriting
        try:
            with open(os.path.join(app.root_path, 'pictures', filename), 'wb') as f:
                f.write(picture.decode('base64'))
        except IOError:
            os.makedirs(os.path.join(app.root_path, 'pictures'))
            with open(os.path.join(app.root_path, 'pictures', filename), 'wb') as f:
                f.write(picture.decode('base64'))
        query_db("INSERT INTO pictures (jid, filename, type, model) \
                 VALUES(?,?,?,?)",
                 [jobid, filename, pic_type, model], insert=True)
        return Response("%s : %s" % (url_for("serve_pic", filename=filename), filename),
                        status=200, mimetype='text/csv')
    except Exception as e:
        return Response("Something went wrong! Python error: %s" % str(e), status=400, mimetype='text/csv')


@app.route('/serve_picture/<filename>', methods=['GET'])
def serve_pic(filename):
    path = os.path.join(app.root_path, 'pictures', filename)
    if os.path.exists(path):
        return send_from_directory(os.path.join(app.root_path, 'pictures'), filename)
    else:
        return Response("Requested file %s does not exist in pictures folder" % filename,
                        status=400, mimetype="text/plain")


@app.route('/delete_picture', methods=['POST'])
def delete_pic():
    filename = request.json['filename']
    path = os.path.join(app.root_path, 'pictures', filename)
    if os.path.exists(path):
        try:
            os.remove(path)
            query_db("DELETE FROM pictures WHERE filename=? ", [filename], insert=True)
            return Response("Success!", status=200, mimetype="text/plain")
        except OSError:
            return Response("Failed to delete the file. Perhaps you don't have necessary permissions?" % filename,
                            status=400, mimetype="text/plain")
    else:
        return Response("Requested file %s does not exist in pictures folder" % filename,
                        status=400, mimetype="text/plain")


@app.route('/get_pictures_links/<jobid>', methods=['GET'])
def get_pics(jobid):
    query = query_db("SELECT * FROM pictures WHERE jid=?", [jobid])
    if query is None:
        return jsonify([])
    results = []
    for row in query:
        results.append({'path': url_for("serve_pic", filename=row['filename']),
                        'model': row['model'], 'type': row['type'], 'filename': row['filename']})
    return jsonify(results)


@app.route('/get_plot/<jobid>', methods=['POST'])
def get_plot(jobid):
    model = request.json['model']
    data_dir = query_db("SELECT working_dir FROM user_queue WHERE \
            jid=?", [jobid], one=True)[0]
    filepath = os.path.join(data_dir, model + ".csv")
    script, div = create_plot(filepath, model)
    results = {'script': script, 'div': div}
    return jsonify(results)



@app.route('/get_mut_plot/<jobid>', methods=['POST'])
def get_mut_plot(jobid):
    data_dir = query_db("SELECT working_dir FROM user_queue WHERE \
            jid=?", [jobid], one=True)[0]
    try:
        script, div = create_mut_plot(data_dir)
        results = {'script': script, 'div': div}
        return jsonify(results)
    except IOError:     # The plot is requested from a project that misses files
        return Response("No plots were generated as the Mutations.csv file is missing",
                        status=400, mimetype="text/plain")

@app.route('/favicon.ico')
def favicon():
    return send_from_directory(os.path.join(app.root_path, 'static', 'img'),
                               'favicon.png', mimetype='image/png')

@app.route('/contact')
def index_contact():
    return render_template('contact.html')


@app.route('/learn_more')
def learn_more():
    return render_template('_learn_more.html')


@app.route('/tutorial_txt')
def tutorial_txt():
    return render_template('tutorial_text.html')


@app.route('/tutorial')
def tutorial():
    return render_template('_tutorial.html')




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
import os
import sys
import glob
from os.path import join, isfile

_setup_script_path = os.path.dirname(os.path.abspath(__file__))
_data_path = os.path.join(_setup_script_path, "aggrescan", "data")
if not os.path.exists(_data_path):
    print "data folder not available from %s. Qutting." % _data_path
    sys.exit(1)

all_files = glob.glob(join("a3d_gui", "*"))
final_list = []
while all_files:
    item = all_files.pop()
    if isfile(item):
        final_list.append(item)
    else:
        for i in glob.glob(join(item, "*")):
            all_files.append(i)


setup(
    name='Aggrescan3D',
    version='1.0.1',
    packages=['aggrescan', 'a3d_gui'],
    url='https://bitbucket.org/lcbio/aggrescan3d',
    license='free for non-commercial users',
    author='Laboratory of Computational Biology',
    description='Aggrescan3D - protein aggregation analysis standalone application',
    install_requires=['numpy', 'matplotlib>=2.0', 'requests'],
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'aggrescan = aggrescan.__main__:run_program',
            'a3d_server = a3d_gui.run_server:run_server'
        ]
    },
    package_data={'aggrescan': [os.path.join('data', '*'),
                                os.path.join('data', 'matrices', '*'),
                                os.path.join('data', 'freesasa-2.0.1', 'src', 'freesasa')],
                  'a3d_gui': final_list}
)



from freesasa import *
import unittest
import math
import os
from exceptions import Exception

# this class tests using derived classes to create custom Classifiers
class DerivedClassifier(Classifier):
    purePython = True

    def classify(self,residueName,atomName):
        return 'bla'

    def radius(self,residueName,atomName):
        return 10

class FreeSASATestCase(unittest.TestCase):
    def testParameters(self):
        d = defaultParameters
        p = Parameters()
        self.assertTrue(p.algorithm() == LeeRichards)
        self.assertTrue(p.algorithm() == d['algorithm'])
        self.assertTrue(p.probeRadius() == d['probe-radius'])
        self.assertTrue(p.nPoints() == d['n-points'])
        self.assertTrue(p.nSlices() == d['n-slices'])
        self.assertTrue(p.nThreads() == d['n-threads'])
        self.assertRaises(AssertionError,lambda: Parameters({'not-an-option' : 1}))
        self.assertRaises(AssertionError,lambda: Parameters({'n-slices' : 50, 'not-an-option' : 1}))
        self.assertRaises(AssertionError,lambda: Parameters({'not-an-option' : 50, 'also-not-an-option' : 1}))

        p.setAlgorithm(ShrakeRupley)
        self.assertTrue(p.algorithm() == ShrakeRupley)
        p.setAlgorithm(LeeRichards)
        self.assertTrue(p.algorithm() == LeeRichards)
        self.assertRaises(AssertionError,lambda: p.setAlgorithm(-10))

        p.setProbeRadius(1.5)
        self.assertTrue(p.probeRadius() == 1.5)
        self.assertRaises(AssertionError,lambda: p.setProbeRadius(-1))

        p.setNPoints(20)
        self.assertTrue(p.nPoints() == 20)
        self.assertRaises(AssertionError,lambda: p.setNPoints(0))

        p.setNSlices(10)
        self.assertTrue(p.nSlices() == 10)
        self.assertRaises(AssertionError,lambda: p.setNSlices(0))

        p.setNThreads(2)
        self.assertTrue(p.nThreads() == 2)
        self.assertRaises(AssertionError, lambda: p.setNThreads(0))
        
    def testResult(self):
        r = Result()
        self.assertRaises(AssertionError,lambda: r.totalArea())
        self.assertRaises(AssertionError,lambda: r.atomArea(0))

    def testClassifier(self):
        c = Classifier()
        self.assertTrue(c._isCClassifier())
        self.assertTrue(c.classify("ALA"," CB ") == apolar)
        self.assertTrue(c.classify("ARG"," NH1") == polar)
        self.assertTrue(c.radius("ALA"," CB ") == 1.88)

        setVerbosity(silent)
        self.assertRaises(Exception,lambda: Classifier("data/err.config"))
        self.assertRaises(IOError,lambda: Classifier(""))
        setVerbosity(normal)

        c = Classifier("data/test.config")
        self.assertTrue(c.classify("AA","aa") == "Polar")
        self.assertTrue(c.classify("BB","bb") == "Apolar")
        self.assertTrue(c.radius("AA","aa") == 1.0)
        self.assertTrue(c.radius("BB","bb") == 2.0)

        c = Classifier("share/oons.config")
        self.assertTrue(c.radius("ALA"," CB ") == 2.00)

        c = DerivedClassifier()
        self.assertTrue(not c._isCClassifier())
        self.assertTrue(c.radius("ALA"," CB ") == 10)
        self.assertTrue(c.radius("ABCDEFG","HIJKLMNO") == 10)
        self.assertTrue(c.classify("ABCDEFG","HIJKLMNO") == "bla")

    def testStructure(self):
        self.assertRaises(IOError,lambda: Structure("xyz#$%"))
        setVerbosity(silent)
        # test any file that's not a PDB file
        self.assertRaises(Exception,lambda: Structure("data/err.config")) 
        self.assertRaises(Exception,lambda: Structure("data/empty.pdb"))
        self.assertRaises(Exception,lambda: Structure("data/empty_model.pdb"))
        setVerbosity(normal)

        s = Structure("data/1ubq.pdb")
        self.assertTrue(s.nAtoms() == 602)
        self.assertTrue(s.radius(1) == 1.88)
        self.assertTrue(s.chainLabel(1) == 'A')
        self.assertTrue(s.atomName(1) == ' CA ')
        self.assertTrue(s.residueName(1) == 'MET')
        self.assertTrue(s.residueNumber(1) == '   1')

        s2 = Structure("data/1ubq.pdb",Classifier("share/oons.config"))
        self.assertTrue(s.nAtoms() == 602)
        self.assertTrue(math.fabs(s2.radius(1) - 2.0) < 1e-5)

        s2 = Structure("data/1ubq.pdb",Classifier("share/protor.config"))
        for i in range (0,601):
            self.assertTrue(math.fabs(s.radius(i)- s2.radius(i)) < 1e-5)

        self.assertRaises(Exception,lambda: Structure("data/1ubq.pdb","data/err.config"))

        s = Structure()
        s.addAtom(' CA ','ALA','   1','A',1,1,1)
        self.assertTrue(s.nAtoms() == 1)
        self.assertTrue(s.atomName(0) == ' CA ')
        self.assertTrue(s.residueName(0) == 'ALA')
        self.assertTrue(s.residueNumber(0) == '   1')
        self.assertTrue(s.chainLabel(0) == 'A')
        self.assertTrue(s.nAtoms() == 1)
        x, y, z = s.coord(0)
        self.assertTrue(x == 1 and y ==1 and z ==1)
        s.addAtom(' CB ','ALA',2,'A',2,1,1)
        self.assertTrue(s.nAtoms() == 2)
        self.assertTrue(s.residueNumber(1) == '2')

        self.assertRaises(AssertionError, lambda: s.atomName(3))
        self.assertRaises(AssertionError, lambda: s.residueName(3))
        self.assertRaises(AssertionError, lambda: s.residueNumber(3))
        self.assertRaises(AssertionError, lambda: s.chainLabel(3))
        self.assertRaises(AssertionError, lambda: s.coord(3))
        self.assertRaises(AssertionError, lambda: s.radius(3))

        s.setRadiiWithClassifier(Classifier())
        self.assertTrue(s.radius(0) == 1.88)
        self.assertTrue(s.radius(1) == 1.88)

        s.setRadiiWithClassifier(DerivedClassifier())
        self.assertTrue(s.radius(0) == s.radius(1) == 10.0)
        
        s.setRadii([1.0,3.0])
        self.assertTrue(s.radius(0) == 1.0)
        self.assertTrue(s.radius(1) == 3.0)

        s.setRadius(0, 10.0)
        self.assertTrue(s.radius(0) == 10.0);

        self.assertRaises(AssertionError,lambda: s.setRadius(2,10));
        self.assertRaises(AssertionError,lambda: s.setRadii([1]))
        self.assertRaises(AssertionError,lambda: s.setRadii([1,2,3]))

        self.assertRaises(AssertionError,lambda: s.atomName(2))
        self.assertRaises(AssertionError,lambda: s.residueName(2))
        self.assertRaises(AssertionError,lambda: s.residueNumber(2))
        self.assertRaises(AssertionError,lambda: s.chainLabel(2))

        setVerbosity(nowarnings)
        s = Structure("data/1d3z.pdb",None,{'hydrogen' : True})
        self.assertTrue(s.nAtoms() == 1231)

        s = Structure("data/1d3z.pdb",None,{'hydrogen' : True, 'join-models' : True})
        self.assertTrue(s.nAtoms() == 12310)

        s = Structure("data/1ubq.pdb",None,{'hetatm' : True})
        self.assertTrue(s.nAtoms() == 660)

        s = Structure("data/1d3z.pdb",None,{'hydrogen' : True, 'skip-unknown' : True})
        self.assertTrue(s.nAtoms() == 602)
        
        setVerbosity(silent)
        self.assertRaises(Exception, lambda : Structure("data/1d3z.pdb",None,{'hydrogen' : True, 'halt-at-unknown' : True}))
        setVerbosity(normal)

    def testStructureArray(self):
        # default separates chains, only uses first model (129 atoms per chain)
        ss = structureArray("data/2jo4.pdb")
        self.assertTrue(len(ss) == 4)
        for s in ss:
            self.assertTrue(s.nAtoms() == 129)

        # include all models, separate chains, and include hydrogen and hetatm (286 atoms per chain)
        setVerbosity(nowarnings)
        ss = structureArray("data/2jo4.pdb",{'separate-models' : True,
                                             'hydrogen' : True,
                                             'hetatm' : True,
                                             'separate-chains' : True})
        self.assertTrue(len(ss) == 4*10)
        for s in ss:
            self.assertTrue(s.nAtoms() == 286)

        # include all models, and include hydrogen and hetatm (286 atoms per chain)
        ss = structureArray("data/2jo4.pdb",{'separate-models' : True,
                                             'hydrogen' : True,
                                             'hetatm' : True})
        self.assertTrue(len(ss) == 10)
        for s in ss:
            self.assertTrue(s.nAtoms() == 286*4)
        setVerbosity(normal)

        # check that the structures initialized this way can be used for calculations
        ss = structureArray("data/1ubq.pdb")
        self.assertTrue(len(ss) == 1)
        self.assertTrue(ss[0].nAtoms() == 602)
        result = calc(ss[0],Parameters({'algorithm' : ShrakeRupley}))
        self.assertTrue(math.fabs(result.totalArea() - 4834.716265) < 1e-5)

        # Test exceptions
        setVerbosity(silent)
        self.assertRaises(AssertionError,lambda: structureArray(None))
        self.assertRaises(IOError,lambda: structureArray(""))
        self.assertRaises(Exception,lambda: structureArray("data/err.config"))
        self.assertRaises(AssertionError,lambda: structureArray("data/2jo4.pdb",{'not-an-option' : True}))
        self.assertRaises(AssertionError,
                          lambda: structureArray("data/2jo4.pdb",
                                                 {'not-an-option' : True, 'hydrogen' : True}))
        self.assertRaises(AssertionError,
                          lambda: structureArray("data/2jo4.pdb",
                                                 {'hydrogen' : True}))
        setVerbosity(normal)

    def testCalc(self):
        # test default settings
        structure = Structure("data/1ubq.pdb")
        result = calc(structure,Parameters({'algorithm' : ShrakeRupley}))
        self.assertTrue(math.fabs(result.totalArea() - 4834.716265) < 1e-5)
        sasa_classes = classifyResults(result,structure)
        self.assertTrue(math.fabs(sasa_classes['Polar'] - 2515.821238) < 1e-5)
        self.assertTrue(math.fabs(sasa_classes['Apolar'] - 2318.895027) < 1e-5)

        # test L&R
        result = calc(structure,Parameters({'algorithm' : LeeRichards, 'n-slices' : 20}))
        sasa_classes = classifyResults(result,structure)
        self.assertTrue(math.fabs(result.totalArea() - 4804.055641) < 1e-5)
        self.assertTrue(math.fabs(sasa_classes['Polar'] - 2504.217302) < 1e-5)
        self.assertTrue(math.fabs(sasa_classes['Apolar'] - 2299.838339) < 1e-5)

        # test extending Classifier with derived class
        sasa_classes = classifyResults(result,structure,DerivedClassifier())
        self.assertTrue(math.fabs(sasa_classes['bla'] - 4804.055641) < 1e-5)
        
        ## test calculating with user-defined classifier ##
        classifier = Classifier("share/oons.config")
        # classifier passed to assign user-defined radii, could also have used setRadiiWithClassifier()
        structure = Structure("data/1ubq.pdb",classifier) 
        result = calc(structure,Parameters({'algorithm' : ShrakeRupley}))
        self.assertTrue(math.fabs(result.totalArea() - 4779.5109924) < 1e-5)
        sasa_classes = classifyResults(result,structure,classifier) # classifier passed to get user-classes
        self.assertTrue(math.fabs(sasa_classes['Polar'] - 2236.9298941) < 1e-5)
        self.assertTrue(math.fabs(sasa_classes['Apolar'] - 2542.5810983) < 1e-5)

    def testCalcCoord(self):
        # one unit sphere
        radii = [1]
        coord = [0,0,0]
        parameters = Parameters()
        parameters.setNSlices(5000)
        parameters.setProbeRadius(0)
        result = calcCoord(coord, radii, parameters)
        self.assertTrue(math.fabs(result.totalArea() - 4*math.pi) < 1e-3)

        # two separate unit spheres
        radii = [1,1]
        coord = [0,0,0, 4,4,4]
        result = calcCoord(coord, radii, parameters)
        self.assertTrue(math.fabs(result.totalArea() - 2*4*math.pi) < 1e-3)

        self.assertRaises(AssertionError,
                          lambda: calcCoord(radii, radii))

    def testSelectArea(self):
        structure = Structure("data/1ubq.pdb")
        result = calc(structure,Parameters({'algorithm' : ShrakeRupley}))
        # will only test that this gets through to the C interface,
        # extensive checking of the parser is done in the C unit tests
        selections = selectArea(('s1, resn ala','s2, resi 1'),structure,result)
        self.assertTrue(math.fabs(selections['s1'] - 118.35) < 0.1)
        self.assertTrue(math.fabs(selections['s2'] - 50.77) < 0.1)

    def testBioPDB(self):
        try:
            from Bio.PDB import PDBParser
        except ImportError:
            print "Can't import Bio.PDB, tests skipped"
            pass
        else:
            parser = PDBParser()
            bp_structure = parser.get_structure("Ubiquitin","data/1ubq.pdb")
            s1 = structureFromBioPDB(bp_structure)
            s2 = Structure("data/1ubq.pdb")
            self.assertTrue(s1.nAtoms() == s2.nAtoms())

            for i in range(0, s2.nAtoms()):
                self.assertTrue(s1.radius(i) == s2.radius(i))
                # there can be tiny errors here
                self.assertTrue(math.fabs(s1.coord(i)[0] - s2.coord(i)[0]) < 1e-5)
                self.assertTrue(math.fabs(s1.coord(i)[1] - s2.coord(i)[1]) < 1e-5)
                self.assertTrue(math.fabs(s1.coord(i)[2] - s2.coord(i)[2]) < 1e-5)

            # because Bio.PDB structures will have slightly different
            # coordinates (due to rounding errors) we set the
            # tolerance as high as 1e-3
            result = calc(s1, Parameters({'algorithm' : LeeRichards, 'n-slices' : 20}))
            print result.totalArea()
            self.assertTrue(math.fabs(result.totalArea() - 4804.055641) < 1e-3)
            sasa_classes = classifyResults(result, s1)
            self.assertTrue(math.fabs(sasa_classes['Polar'] - 2504.217302) < 1e-3)
            self.assertTrue(math.fabs(sasa_classes['Apolar'] - 2299.838339) < 1e-3)

            result, sasa_classes = calcBioPDB(bp_structure, Parameters({'algorithm' : ShrakeRupley}))
            self.assertTrue(math.fabs(result.totalArea() - 4834.716265) < 1e-3)
            self.assertTrue(math.fabs(sasa_classes['Polar'] - 2515.821238) < 1e-3)
            self.assertTrue(math.fabs(sasa_classes['Apolar'] - 2318.895027) < 1e-3)
            print result.totalArea()


if __name__ == '__main__':
    # make sure we're in the right directory (if script is called from
    # outside the directory)
    abspath = os.path.abspath(__file__)
    dirname = os.path.dirname(abspath)
    os.chdir(dirname)
    unittest.main()




from distutils.core import setup
from distutils.extension import Extension
from Cython.Build import cythonize

lib = []
lib.extend([]);
lib.extend([]);

extensions = [
    Extension("*", ["*.pyx"],
              include_dirs = ["../../src"],
              language='c',
              libraries=lib,
              extra_objects = ["../../src/libfreesasa.a"],
              extra_compile_args = ["-w"] 
              )
]

setup(
    name='FreeSASA',
    description='Calculate solvent accessible surface areas of proteins',
    version= '2.0.1',
    author='Simon Mitternacht',
    url='http://freesasa.github.io/',
    license='MIT',
    ext_modules = cythonize(extensions)
)



# -*- coding: utf-8 -*-

import os
import shutil
import logger
from subprocess import PIPE, Popen
from os.path import join, isfile
import platform


_name = "FoldX"
if platform.system() == "Windows":
    _bin_name = "foldx.exe"
else:
    _bin_name = "foldx"
# TODO once this is working again re-write given the fact that command line rather than files can be now used
# TODO maybe exploit the new foldx options for a cleaner pipeline of the pdb files etc


class FoldxWrap:
    """
    Expects a path to folder with FoldX binary and rotabase.txt files and a script dir with repairPDB.txt file
    If the file is not present in scripts dir, the class will attempt to create it
    This class can be used to run energy minimization and mutant creation with foldx
    """
    def __init__(self, foldx_dir="", script_dir="", work_dir="", skip_minimization=False):
        self.bin_dir = foldx_dir
        self.script_dir = script_dir
        self.fold_config = join(self.script_dir, "repairPDB.txt")
        self.mut_fold_config = join(self.script_dir, "buildModel.txt")
        self.agg_work_dir = work_dir
        self.skip = skip_minimization
        try:
            self._check_files()
        except logger.FoldXError:
            raise

    def minimize_energy(self, working_dir="", cleanup=True, skip_min=False):
        try:
            os.chdir(working_dir)
        except OSError:
            raise logger.FoldXError("FoldX run in a non-existing directory %s" % working_dir)
        if not isfile(join(working_dir, "input.pdb")):
            raise logger.FoldXError("input.pdb file not present if FoldX working directory %s" % working_dir)
        if not isfile(join(working_dir, "rotabase.txt")):
            # In case of multiple minimizations te code below could raise a same file error
            try:
                os.symlink(join(self.bin_dir, "rotabase.txt"), join(working_dir, "rotabase.txt"))
                os.symlink(join(self.script_dir, "repairPDB.txt"), join(working_dir, "repairPDB.txt"))
            except (OSError, AttributeError):   # Windows will not necessarily allow symlink creation
                shutil.copyfile(join(self.bin_dir, "rotabase.txt"), join(working_dir, "rotabase.txt"))
                shutil.copyfile(join(self.script_dir, "repairPDB.txt"), join(working_dir, "repairPDB.txt"))
        fold_cmd = [join(self.bin_dir, _bin_name), "-f", "repairPDB.txt"]
        if not skip_min:
            self._run_foldx(cmd=fold_cmd, msg="Starting FoldX energy minimalization", outfile="finalRepair.out")
        else:
            # since the job is ran on output from parent this "input" should be the output of the parent process
            shutil.move("input.pdb", "input_Repair.pdb")    # Pretend it was ran
        if cleanup:
            try:
                shutil.move("input_Repair.fxout", "RepairPDB.fxout")
                shutil.move("input_Repair.pdb", "folded.pdb")
            except IOError:
                raise logger.FoldXError("FoldX didn't produce expected repair files. "
                                        "Can't continue without it. This is unexpected and could indicate FoldX issues.")

    def build_mutant(self, working_dir="", mutation_list=""):
        os.chdir(working_dir)
        logger.to_file(filename="individual_list.txt", content=mutation_list)
        self.minimize_energy(working_dir=working_dir, cleanup=False, skip_min=self.skip)
        _cmd = [join(self.bin_dir, _bin_name), "-f", self.mut_fold_config]
        self._run_foldx(cmd=_cmd, msg="Building mutant model", outfile="buildMutant.out")
        try:
            shutil.move("input_Repair_1.pdb", "input.pdb")
        except IOError:
            raise logger.FoldXError("FoldX didn't produce expected mutant file. "
                                    "Can't continue without it. This is unexpected and could indicate FoldX issues.")
        self._cleanup()

    @staticmethod
    def _run_foldx(cmd, msg='', outfile=''):
        """Assumes the directory it's run from contains """
        logger.info(module_name="FoldX", msg=msg)
        logger.debug(module_name="FoldX", msg="Starting FoldX with %s" % " ".join(cmd))
        out, err = Popen(cmd, stdout=PIPE, stderr=PIPE).communicate()
        if err:
            logger.warning(module_name="FoldX", msg=err)
        logger.to_file(filename=outfile, content=out, allow_err=False)

# TODO proper cleanup handling this is a mess + is this even necessary?
    def _cleanup(self):
        with open("Dif_input_Repair.fxout", 'r') as enelog:
            d = enelog.readlines()[-1]
            d = d.split()[1] + " kcal/mol"
            logger.to_file(filename=join(self.agg_work_dir, "MutantEnergyDiff"), content=d)

        try:
            os.mkdir(join(self.agg_work_dir, "foldxOutput"))
        except OSError:
            logger.log_file(module_name=_name, msg="Couldn't create foldxOutput folder (probably already exists)")
        shutil.move("Raw_input_Repair.fxout",
                    join(self.agg_work_dir, "foldxOutput/Stability_RepairPDB_input.fxout"))
        shutil.move("Dif_input_Repair.fxout",
                    join(self.agg_work_dir, "foldxOutput/Dif_BuildModel_RepairPDB_input.fxout"))

    def _check_files(self):
        if not isfile(join(self.bin_dir,"rotabase.txt")) or not isfile(join(self.bin_dir,_bin_name)):
            raise logger.FoldXError("Provided FoldX directory (%s) misses either %s or %s " %
                                    (self.bin_dir, "rotabase.txt", _bin_name) +
                                    "files which are required for the program to run.")
        if not isfile(self.fold_config):
            try:
                self._recover_script()
            except logger.FoldXError:
                raise

        if not isfile(self.mut_fold_config):
            try:
                self._recover_mut_script()
            except logger.FoldXError:
                raise

    def _recover_script(self):
        """
        Re-write the repairPDB file in case something goes wrong
        For more info on FoldX commands and parameters see http://foldxsuite.crg.eu/documentation#manual
        Default FoldX parameters include:
        ph=7
        temperature=298
        ionStrength=0.05
        vdwDesign=2
        pdbHydrogens=false
        """
        fold_script = (
            "pdb=input.pdb\n"
            "command=RepairPDB\n"
            "water=-CRYSTAL\n"
            )
        try:
            with open(self.fold_config, "w") as f:
                f.write(fold_script)
        except IOError:
            raise logger.FoldXError("FoldX requires %s file to be present." % self.fold_config +
                                    "If default settings are used it should be in data folder inside aggrescan directory." +
                                    "Attempted to re-write the file %s to %s but failed" % (self.fold_config, self.script_dir))

    def _recover_mut_script(self):
        """
        Re-write the buildModel file in case something goes wrong
        For more info on FoldX commands and parameters see http://foldxsuite.crg.eu/documentation#manual
        Default FoldX parameters include:
        ph=7
        temperature=298
        ionStrength=0.05
        vdwDesign=2
        pdbHydrogens=false
        """
        fold_mut_script = (
            "pdb=input_Repair.pdb\n"
            "command=RepairPDB\n"
            "water=-CRYSTAL\n")
        try:
            with open(self.mut_fold_config, "w") as f:
                f.write(fold_mut_script)
        except IOError:
            raise logger.FoldXError("FoldX mutation requires %s file to be present." % self.mut_fold_config +
                                    "If default settings are used it should be in data folder inside aggrescan directory." +
                                    "Attempted to re-write the file %s to %s but failed" % (self.mut_fold_config, self.script_dir))

